from types import MappingProxyType
from typing import Dict
from typing import Mapping
from typing import Tuple

import attrs
from typeguard import typechecked

from tecton_core import id_helper
from tecton_core.specs import feature_view_spec
from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_proto.args import feature_service__client_pb2 as feature_service__args_pb2
from tecton_proto.data import feature_service__client_pb2 as feature_service__data_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


__all__ = [
    "FeatureServiceSpec",
    "FeatureSetItemSpec",
    "FeatureServiceSpecArgsSupplement",
]


@utils.frozen_strict
class FeatureServiceSpec(tecton_object_spec.TectonObjectSpec):
    feature_set_items: Tuple["FeatureSetItemSpec", ...]
    online_serving_enabled: bool
    options: Mapping[str, str]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: feature_service__data_pb2.FeatureService) -> "FeatureServiceSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.feature_service_id, proto.fco_metadata
            ),
            feature_set_items=tuple(FeatureSetItemSpec.from_data_proto(item) for item in proto.feature_set_items),
            online_serving_enabled=proto.online_serving_enabled,
            validation_args=validator_pb2.FcoValidationArgs(feature_service=proto.validation_args),
            options=MappingProxyType(proto.options),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: feature_service__args_pb2.FeatureServiceArgs, supplement: "FeatureServiceSpecArgsSupplement"
    ) -> "FeatureServiceSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(proto.feature_service_id, proto.info),
            feature_set_items=tuple(
                FeatureSetItemSpec.from_args_proto(fp, supplement) for fp in proto.feature_references
            ),
            online_serving_enabled=proto.online_serving_enabled,
            validation_args=None,
            options=MappingProxyType(proto.options),
        )


@attrs.define
class FeatureServiceSpecArgsSupplement:
    """A data class used for supplementing args protos during FeatureServiceSpec construction.

    This Python data class can be used to include data that is not included in args protos (e.g. schemas) into the
    FeatureServiceSpec constructor.
    """

    ids_to_feature_views: Dict[str, feature_view_spec.FeatureViewSpec]


@utils.frozen_strict
class FeatureSetItemSpec:
    feature_view_id: str
    namespace: str
    feature_columns: Tuple[str, ...]
    # Mapping from spine join key to the Feature View join keys.  Not a dict to account for multi-mapping.
    join_key_mappings: Tuple["utils.JoinKeyMappingSpec", ...]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: feature_service__data_pb2.FeatureSetItem) -> "FeatureSetItemSpec":
        join_key_mappings = []
        for join_configuration_item in proto.join_configuration_items:
            join_key_mappings.append(
                utils.JoinKeyMappingSpec(
                    spine_column_name=join_configuration_item.spine_column_name,
                    feature_view_column_name=join_configuration_item.package_column_name,
                )
            )

        return cls(
            feature_view_id=id_helper.IdHelper.to_string(proto.feature_view_id),
            namespace=utils.get_field_or_none(proto, "namespace"),
            feature_columns=utils.get_tuple_from_repeated_field(proto.feature_columns),
            join_key_mappings=tuple(join_key_mappings),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: feature_service__args_pb2.FeatureReference, supplement: FeatureServiceSpecArgsSupplement
    ) -> "FeatureSetItemSpec":
        feature_view_id = id_helper.IdHelper.to_string(proto.feature_view_id)
        fv_spec = supplement.ids_to_feature_views[feature_view_id]

        override_mapping = {
            column_pair.feature_column: column_pair.spine_column for column_pair in proto.override_join_keys
        }
        join_key_mappings = []
        for join_key in fv_spec.join_keys:
            if join_key in override_mapping:
                join_key_mappings.append(
                    utils.JoinKeyMappingSpec(
                        spine_column_name=override_mapping[join_key],
                        feature_view_column_name=join_key,
                    )
                )
            else:
                join_key_mappings.append(
                    utils.JoinKeyMappingSpec(
                        spine_column_name=join_key,
                        feature_view_column_name=join_key,
                    )
                )

        if len(proto.features) > 0:
            # Use a subset of the features from the feature view.
            features = utils.get_tuple_from_repeated_field(proto.features)
        else:
            features = tuple(fv_spec.features)

        return cls(
            feature_view_id=feature_view_id,
            namespace=utils.get_field_or_none(proto, "namespace"),
            feature_columns=features,
            join_key_mappings=tuple(join_key_mappings),
        )


# Resolve forward type declarations.
attrs.resolve_types(FeatureServiceSpec, locals(), globals())
attrs.resolve_types(FeatureSetItemSpec, locals(), globals())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/testhelperservice/test_helper_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n@tecton_proto/testhelperservice/test_helper_service__client.proto\x12\x1etecton_proto.testhelperservice\x1a\'tecton_proto/auth/service__client.proto\"K\n\x1cIngestionServerConfigRequest\x12+\n\x11\x61\x62solute_filepath\x18\x01 \x01(\tR\x10\x61\x62soluteFilepath\"\x1f\n\x1dIngestionServerConfigResponse\"b\n\x16InitializeStoreRequest\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12*\n\x11\x66\x65\x61ture_view_name\x18\x02 \x01(\tR\x0f\x66\x65\x61tureViewName\"\x19\n\x17InitializeStoreResponse2\xf5\x02\n\x11TestHelperService\x12\xb7\x01\n\x15IngestionServerConfig\x12<.tecton_proto.testhelperservice.IngestionServerConfigRequest\x1a=.tecton_proto.testhelperservice.IngestionServerConfigResponse\"!\xa2\xbc\xe6\xc0\x05\x1b*\x19tecton/organization#admin\x12\xa5\x01\n\x0fInitializeStore\x12\x36.tecton_proto.testhelperservice.InitializeStoreRequest\x1a\x37.tecton_proto.testhelperservice.InitializeStoreResponse\"!\xa2\xbc\xe6\xc0\x05\x1b*\x19tecton/organization#adminBm\n\x1c\x63om.tecton.testhelperserviceB\x16TestHelperServiceProtoP\x01Z3github.com/tecton-ai/tecton_proto/testhelperservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.testhelperservice.test_helper_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.tecton.testhelperserviceB\026TestHelperServiceProtoP\001Z3github.com/tecton-ai/tecton_proto/testhelperservice'
  _TESTHELPERSERVICE.methods_by_name['IngestionServerConfig']._options = None
  _TESTHELPERSERVICE.methods_by_name['IngestionServerConfig']._serialized_options = b'\242\274\346\300\005\033*\031tecton/organization#admin'
  _TESTHELPERSERVICE.methods_by_name['InitializeStore']._options = None
  _TESTHELPERSERVICE.methods_by_name['InitializeStore']._serialized_options = b'\242\274\346\300\005\033*\031tecton/organization#admin'
  _INGESTIONSERVERCONFIGREQUEST._serialized_start=141
  _INGESTIONSERVERCONFIGREQUEST._serialized_end=216
  _INGESTIONSERVERCONFIGRESPONSE._serialized_start=218
  _INGESTIONSERVERCONFIGRESPONSE._serialized_end=249
  _INITIALIZESTOREREQUEST._serialized_start=251
  _INITIALIZESTOREREQUEST._serialized_end=349
  _INITIALIZESTORERESPONSE._serialized_start=351
  _INITIALIZESTORERESPONSE._serialized_end=376
  _TESTHELPERSERVICE._serialized_start=379
  _TESTHELPERSERVICE._serialized_end=752
# @@protoc_insertion_point(module_scope)



from typeguard import typechecked

from tecton import version as tecton_version
from tecton._internals import metadata_service
from tecton.framework import base_tecton_object
from tecton_core import errors
from tecton_core import id_helper
from tecton_proto.data import state_update__client_pb2 as state_update_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


_INDENTATION_SIZE = 4


def _get_validation_prefix(obj: base_tecton_object.BaseTectonObject, indentation_level: int) -> str:
    indent_prefix = " " * (indentation_level * _INDENTATION_SIZE)
    return f"{indent_prefix}{obj.__class__.__name__} '{obj.name}': "


def print_validation_success(
    obj: base_tecton_object.BaseTectonObject,
    indentation_level: int,
) -> None:
    prefix = _get_validation_prefix(obj, indentation_level)
    # TODO(TEC-14146): Use the logging module instead of print statements.
    print(f"{prefix}Successfully validated.")


def format_server_errors(
    tecton_object: base_tecton_object.BaseTectonObject, response: metadata_service_pb2.ValidateLocalFcoResponse
) -> str:
    # If there's a server side error, print that instead of the validation errors.
    if response.error:
        return f"{tecton_object.__class__.__name__} '{tecton_object.info.name}' failed validation: {response.error}"

    validation_errors = response.validation_result.errors
    if len(validation_errors) == 1:
        # If there is a single error, print the entire exception on one line. This is better UX in notebooks, which
        # often only show the first line of an exception in a preview.
        error = validation_errors[0]
        _check_error_matches_tecton_object(tecton_object, error)
        return f"{tecton_object.__class__.__name__} '{tecton_object.info.name}' failed validation: {error.message}"
    else:
        error_strings = [
            f"{tecton_object.__class__.__name__} '{tecton_object.info.name}' failed validation with the following errors:"
        ]
        for error in validation_errors:
            _check_error_matches_tecton_object(tecton_object, error)
            error_strings.append(error.message)
        return "\n".join(error_strings)


def _check_error_matches_tecton_object(
    tecton_object: base_tecton_object.BaseTectonObject, error: state_update_pb2.ValidationMessage
) -> None:
    error_object_id = id_helper.IdHelper.to_string(error.fco_refs[0].fco_id)

    if error_object_id != tecton_object.id:
        msg = f"Backend validation error returned wrong object id: {error_object_id}. Expected {tecton_object.id}"
        raise errors.TectonInternalError(msg)


@typechecked
def run_backend_validation_and_assert_valid(
    tecton_object: base_tecton_object.BaseTectonObject,
    validation_request: validator_pb2.ValidationRequest,
    indentation_level: int,
) -> None:
    """Run validation against the Tecton backend.

    Raises an exception if validation fails. Prints message if successful.
    """
    validation_local_fco_request = metadata_service_pb2.ValidateLocalFcoRequest(
        sdk_version=tecton_version.get_semantic_version(),
        validation_request=validation_request,
    )
    response = metadata_service.instance().ValidateLocalFco(validation_local_fco_request).response_proto
    if response.success:
        print_validation_success(tecton_object, indentation_level)
    else:
        raise errors.TectonValidationError(format_server_errors(tecton_object, response))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/remote_spark__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import data_source_config__client_pb2 as tecton__proto_dot_args_dot_data__source__config____client__pb2
from tecton_proto.args import feature_view__client_pb2 as tecton__proto_dot_args_dot_feature__view____client__pb2
from tecton_proto.args import user_defined_function__client_pb2 as tecton__proto_dot_args_dot_user__defined__function____client__pb2
from tecton_proto.args import virtual_data_source__client_pb2 as tecton__proto_dot_args_dot_virtual__data__source____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.common import spark_schema__client_pb2 as tecton__proto_dot_common_dot_spark__schema____client__pb2
from tecton_proto.data import feature_view__client_pb2 as tecton__proto_dot_data_dot_feature__view____client__pb2
from tecton_proto.data import hive_metastore__client_pb2 as tecton__proto_dot_data_dot_hive__metastore____client__pb2
from tecton_proto.data import transformation__client_pb2 as tecton__proto_dot_data_dot_transformation____client__pb2
from tecton_proto.data import virtual_data_source__client_pb2 as tecton__proto_dot_data_dot_virtual__data__source____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,tecton_proto/data/remote_spark__client.proto\x12\x11tecton_proto.data\x1a\x32tecton_proto/args/data_source_config__client.proto\x1a,tecton_proto/args/feature_view__client.proto\x1a\x35tecton_proto/args/user_defined_function__client.proto\x1a\x33tecton_proto/args/virtual_data_source__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a.tecton_proto/common/spark_schema__client.proto\x1a,tecton_proto/data/feature_view__client.proto\x1a.tecton_proto/data/hive_metastore__client.proto\x1a.tecton_proto/data/transformation__client.proto\x1a\x33tecton_proto/data/virtual_data_source__client.proto\"\xf2\x01\n\x12GetHiveTableSchema\x12\x1a\n\x08\x64\x61tabase\x18\x01 \x01(\tR\x08\x64\x61tabase\x12\x14\n\x05table\x18\x02 \x01(\tR\x05table\x12(\n\x0ftimestampColumn\x18\x03 \x01(\tR\x0ftimestampColumn\x12(\n\x0ftimestampFormat\x18\x04 \x01(\tR\x0ftimestampFormat\x12V\n\x12rawBatchTranslator\x18\x05 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x12rawBatchTranslator\"\x89\x02\n\x13GetUnityTableSchema\x12\x18\n\x07\x63\x61talog\x18\x01 \x01(\tR\x07\x63\x61talog\x12\x16\n\x06schema\x18\x02 \x01(\tR\x06schema\x12\x14\n\x05table\x18\x03 \x01(\tR\x05table\x12(\n\x0ftimestampColumn\x18\x04 \x01(\tR\x0ftimestampColumn\x12(\n\x0ftimestampFormat\x18\x05 \x01(\tR\x0ftimestampFormat\x12V\n\x12rawBatchTranslator\x18\x06 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x12rawBatchTranslator\"\x9e\x01\n GetBatchDataSourceFunctionSchema\x12\x42\n\x08\x66unction\x18\x01 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x08\x66unction\x12\x36\n\x17supports_time_filtering\x18\x02 \x01(\x08R\x15supportsTimeFiltering\"g\n!GetStreamDataSourceFunctionSchema\x12\x42\n\x08\x66unction\x18\x01 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x08\x66unction\"\xd8\x01\n\x16GetRedshiftTableSchema\x12\x1a\n\x08\x65ndpoint\x18\x01 \x01(\tR\x08\x65ndpoint\x12\x14\n\x05table\x18\x02 \x01(\tR\x05table\x12\x14\n\x05query\x18\x03 \x01(\tR\x05query\x12V\n\x12rawBatchTranslator\x18\x04 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x12rawBatchTranslator\x12\x1e\n\x0btemp_s3_dir\x18\x05 \x01(\tR\ttempS3Dir\"\x95\x02\n\x12GetSnowflakeSchema\x12\x10\n\x03url\x18\x01 \x01(\tR\x03url\x12\x12\n\x04role\x18\x02 \x01(\tR\x04role\x12\x1a\n\x08\x64\x61tabase\x18\x03 \x01(\tR\x08\x64\x61tabase\x12\x16\n\x06schema\x18\x04 \x01(\tR\x06schema\x12\x1c\n\twarehouse\x18\x05 \x01(\tR\twarehouse\x12\x16\n\x05table\x18\x06 \x01(\tH\x00R\x05table\x12\x16\n\x05query\x18\x07 \x01(\tH\x00R\x05query\x12M\n\x0epost_processor\x18\x08 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\rpostProcessorB\x08\n\x06source\"\x8d\x03\n\x13GetFileSourceSchema\x12\x10\n\x03uri\x18\x01 \x01(\tR\x03uri\x12\x1e\n\nfileFormat\x18\x02 \x01(\tR\nfileFormat\x12\x30\n\x13\x63onvertToGlueFormat\x18\x03 \x01(\x08R\x13\x63onvertToGlueFormat\x12V\n\x12rawBatchTranslator\x18\x04 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x12rawBatchTranslator\x12\x1c\n\tschemaUri\x18\x05 \x01(\tR\tschemaUri\x12(\n\x0ftimestampColumn\x18\x06 \x01(\tR\x0ftimestampColumn\x12(\n\x0ftimestampFormat\x18\x07 \x01(\tR\x0ftimestampFormat\x12H\n\x0eschemaOverride\x18\x08 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0eschemaOverride\"\x92\x01\n\x16GetKinesisSourceSchema\x12\x1e\n\nstreamName\x18\x01 \x01(\tR\nstreamName\x12X\n\x13rawStreamTranslator\x18\x02 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x13rawStreamTranslator\"\xea\x01\n\x14GetKafkaSourceSchema\x12X\n\x13rawStreamTranslator\x18\x01 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x13rawStreamTranslator\x12\x32\n\x15ssl_keystore_location\x18\x02 \x01(\tR\x13sslKeystoreLocation\x12\x44\n\x1fssl_keystore_password_secret_id\x18\x03 \x01(\tR\x1bsslKeystorePasswordSecretId\"\x8e\x02\n\x14GetFeatureViewSchema\x12V\n\x14virtual_data_sources\x18\x02 \x03(\x0b\x32$.tecton_proto.data.VirtualDataSourceR\x12virtualDataSources\x12K\n\x0ftransformations\x18\x03 \x03(\x0b\x32!.tecton_proto.data.TransformationR\x0ftransformations\x12\x45\n\x0c\x66\x65\x61ture_view\x18\x05 \x01(\x0b\x32\".tecton_proto.args.FeatureViewArgsR\x0b\x66\x65\x61tureViewJ\x04\x08\x01\x10\x02J\x04\x08\x04\x10\x05\"\x9a\x02\n&GetQueryPlanInfoForFeatureViewPipeline\x12V\n\x14virtual_data_sources\x18\x02 \x03(\x0b\x32$.tecton_proto.data.VirtualDataSourceR\x12virtualDataSources\x12K\n\x0ftransformations\x18\x03 \x03(\x0b\x32!.tecton_proto.data.TransformationR\x0ftransformations\x12\x45\n\x0c\x66\x65\x61ture_view\x18\x04 \x01(\x0b\x32\".tecton_proto.args.FeatureViewArgsR\x0b\x66\x65\x61tureViewJ\x04\x08\x01\x10\x02\"\x13\n\x11ListHiveDatabases\",\n\x0eListHiveTables\x12\x1a\n\x08\x64\x61tabase\x18\x01 \x01(\tR\x08\x64\x61tabase\"H\n\x14ListHiveTableColumns\x12\x1a\n\x08\x64\x61tabase\x18\x01 \x01(\tR\x08\x64\x61tabase\x12\x14\n\x05table\x18\x02 \x01(\tR\x05table\"\xd3\x01\n\x18\x46\x65\x61tureViewSchemaRequest\x12\x45\n\x0c\x66\x65\x61ture_view\x18\x01 \x01(\x0b\x32\".tecton_proto.args.FeatureViewArgsR\x0b\x66\x65\x61tureView\x12\x1b\n\tjoin_keys\x18\x02 \x03(\tR\x08joinKeys\x12S\n\x12temporal_aggregate\x18\x03 \x01(\x0b\x32$.tecton_proto.data.TemporalAggregateR\x11temporalAggregate\"\xac\x02\n$GetMultipleFeatureViewSchemasRequest\x12V\n\x14virtual_data_sources\x18\x01 \x03(\x0b\x32$.tecton_proto.data.VirtualDataSourceR\x12virtualDataSources\x12K\n\x0ftransformations\x18\x02 \x03(\x0b\x32!.tecton_proto.data.TransformationR\x0ftransformations\x12_\n\x15\x66\x65\x61ture_view_requests\x18\x03 \x03(\x0b\x32+.tecton_proto.data.FeatureViewSchemaRequestR\x13\x66\x65\x61tureViewRequests\"\xa9\x02\n%GetMultipleFeatureViewSchemasResponse\x12\x88\x01\n\x16\x66\x65\x61ture_view_responses\x18\x01 \x03(\x0b\x32R.tecton_proto.data.GetMultipleFeatureViewSchemasResponse.FeatureViewResponsesEntryR\x14\x66\x65\x61tureViewResponses\x1au\n\x19\x46\x65\x61tureViewResponsesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x42\n\x05value\x18\x02 \x01(\x0b\x32,.tecton_proto.data.FeatureViewSchemaResponseR\x05value:\x02\x38\x01\"y\n#GetMultipleDataSourceSchemasRequest\x12R\n\x10\x64\x61ta_source_args\x18\x01 \x03(\x0b\x32(.tecton_proto.args.VirtualDataSourceArgsR\x0e\x64\x61taSourceArgs\"\x96\x02\n$GetMultipleDataSourceSchemasResponse\x12\x84\x01\n\x15\x64\x61ta_source_responses\x18\x01 \x03(\x0b\x32P.tecton_proto.data.GetMultipleDataSourceSchemasResponse.DataSourceResponsesEntryR\x13\x64\x61taSourceResponses\x1ag\n\x18\x44\x61taSourceResponsesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x35\n\x05value\x18\x02 \x01(\x0b\x32\x1f.tecton_proto.data.SparkSchemasR\x05value:\x02\x38\x01\"\x98\x01\n\x0cSparkSchemas\x12\x42\n\x0b\x62\x61tchSchema\x18\x01 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0b\x62\x61tchSchema\x12\x44\n\x0cstreamSchema\x18\x02 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0cstreamSchema\"~\n\x19\x46\x65\x61tureViewSchemaResponse\x12<\n\x0bview_schema\x18\x01 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\nviewSchema\x12#\n\rtimestamp_key\x18\x03 \x01(\tR\x0ctimestampKey\"\xfd\x0e\n\x0e\x45xecuteRequest\x12\x8d\x01\n$getMultipleFeatureViewSchemasRequest\x18\x18 \x01(\x0b\x32\x37.tecton_proto.data.GetMultipleFeatureViewSchemasRequestH\x00R$getMultipleFeatureViewSchemasRequest\x12\x8a\x01\n#getMultipleDataSourceSchemasRequest\x18\x19 \x01(\x0b\x32\x36.tecton_proto.data.GetMultipleDataSourceSchemasRequestH\x00R#getMultipleDataSourceSchemasRequest\x12W\n\x12getHiveTableSchema\x18\x01 \x01(\x0b\x32%.tecton_proto.data.GetHiveTableSchemaH\x00R\x12getHiveTableSchema\x12\x63\n\x16getRedshiftTableSchema\x18\x02 \x01(\x0b\x32).tecton_proto.data.GetRedshiftTableSchemaH\x00R\x16getRedshiftTableSchema\x12Z\n\x13getFileSourceSchema\x18\x03 \x01(\x0b\x32&.tecton_proto.data.GetFileSourceSchemaH\x00R\x13getFileSourceSchema\x12\x63\n\x16getKinesisSourceSchema\x18\x04 \x01(\x0b\x32).tecton_proto.data.GetKinesisSourceSchemaH\x00R\x16getKinesisSourceSchema\x12]\n\x14getKafkaSourceSchema\x18\x05 \x01(\x0b\x32\'.tecton_proto.data.GetKafkaSourceSchemaH\x00R\x14getKafkaSourceSchema\x12]\n\x14getFeatureViewSchema\x18\x0e \x01(\x0b\x32\'.tecton_proto.data.GetFeatureViewSchemaH\x00R\x14getFeatureViewSchema\x12W\n\x12getSnowflakeSchema\x18\x0c \x01(\x0b\x32%.tecton_proto.data.GetSnowflakeSchemaH\x00R\x12getSnowflakeSchema\x12\x81\x01\n getBatchDataSourceFunctionSchema\x18\x15 \x01(\x0b\x32\x33.tecton_proto.data.GetBatchDataSourceFunctionSchemaH\x00R getBatchDataSourceFunctionSchema\x12\x84\x01\n!getStreamDataSourceFunctionSchema\x18\x16 \x01(\x0b\x32\x34.tecton_proto.data.GetStreamDataSourceFunctionSchemaH\x00R!getStreamDataSourceFunctionSchema\x12Z\n\x13getUnityTableSchema\x18\x17 \x01(\x0b\x32&.tecton_proto.data.GetUnityTableSchemaH\x00R\x13getUnityTableSchema\x12T\n\x11listHiveDatabases\x18\x12 \x01(\x0b\x32$.tecton_proto.data.ListHiveDatabasesH\x00R\x11listHiveDatabases\x12K\n\x0elistHiveTables\x18\x13 \x01(\x0b\x32!.tecton_proto.data.ListHiveTablesH\x00R\x0elistHiveTables\x12]\n\x14listHiveTableColumns\x18\x14 \x01(\x0b\x32\'.tecton_proto.data.ListHiveTableColumnsH\x00R\x14listHiveTableColumns\x12\x93\x01\n&getQueryPlanInfoForFeatureViewPipeline\x18\r \x01(\x0b\x32\x39.tecton_proto.data.GetQueryPlanInfoForFeatureViewPipelineH\x00R&getQueryPlanInfoForFeatureViewPipeline\x12H\n\x07\x65nvVars\x18\x10 \x03(\x0b\x32..tecton_proto.data.ExecuteRequest.EnvVarsEntryR\x07\x65nvVars\x1a:\n\x0c\x45nvVarsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\t\n\x07requestJ\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\tJ\x04\x08\t\x10\nJ\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0cJ\x04\x08\x11\x10\x12\"]\n\rQueryPlanInfo\x12\x1b\n\thas_joins\x18\x01 \x01(\x08R\x08hasJoins\x12)\n\x10has_aggregations\x18\x02 \x01(\x08R\x0fhasAggregationsJ\x04\x08\x03\x10\x04\"\x94\x05\n\rExecuteResult\x12&\n\runcaughtError\x18\x01 \x01(\tH\x00R\runcaughtError\x12*\n\x0fvalidationError\x18\x05 \x01(\tH\x00R\x0fvalidationError\x12\x44\n\x0bsparkSchema\x18\x02 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaH\x00R\x0bsparkSchema\x12H\n\rqueryPlanInfo\x18\x03 \x01(\x0b\x32 .tecton_proto.data.QueryPlanInfoH\x00R\rqueryPlanInfo\x12\x35\n\x06schema\x18\x04 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaH\x00R\x06schema\x12K\n\x0elistHiveResult\x18\x06 \x01(\x0b\x32!.tecton_proto.data.ListHiveResultH\x00R\x0elistHiveResult\x12\x88\x01\n!multipleFeatureViewSchemaResponse\x18\x08 \x01(\x0b\x32\x38.tecton_proto.data.GetMultipleFeatureViewSchemasResponseH\x00R!multipleFeatureViewSchemaResponse\x12\x85\x01\n multipleDataSourceSchemaResponse\x18\t \x01(\x0b\x32\x37.tecton_proto.data.GetMultipleDataSourceSchemasResponseH\x00R multipleDataSourceSchemaResponseB\x08\n\x06resultB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.remote_spark__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _GETMULTIPLEFEATUREVIEWSCHEMASRESPONSE_FEATUREVIEWRESPONSESENTRY._options = None
  _GETMULTIPLEFEATUREVIEWSCHEMASRESPONSE_FEATUREVIEWRESPONSESENTRY._serialized_options = b'8\001'
  _GETMULTIPLEDATASOURCESCHEMASRESPONSE_DATASOURCERESPONSESENTRY._options = None
  _GETMULTIPLEDATASOURCESCHEMASRESPONSE_DATASOURCERESPONSESENTRY._serialized_options = b'8\001'
  _EXECUTEREQUEST_ENVVARSENTRY._options = None
  _EXECUTEREQUEST_ENVVARSENTRY._serialized_options = b'8\001'
  _GETHIVETABLESCHEMA._serialized_start=559
  _GETHIVETABLESCHEMA._serialized_end=801
  _GETUNITYTABLESCHEMA._serialized_start=804
  _GETUNITYTABLESCHEMA._serialized_end=1069
  _GETBATCHDATASOURCEFUNCTIONSCHEMA._serialized_start=1072
  _GETBATCHDATASOURCEFUNCTIONSCHEMA._serialized_end=1230
  _GETSTREAMDATASOURCEFUNCTIONSCHEMA._serialized_start=1232
  _GETSTREAMDATASOURCEFUNCTIONSCHEMA._serialized_end=1335
  _GETREDSHIFTTABLESCHEMA._serialized_start=1338
  _GETREDSHIFTTABLESCHEMA._serialized_end=1554
  _GETSNOWFLAKESCHEMA._serialized_start=1557
  _GETSNOWFLAKESCHEMA._serialized_end=1834
  _GETFILESOURCESCHEMA._serialized_start=1837
  _GETFILESOURCESCHEMA._serialized_end=2234
  _GETKINESISSOURCESCHEMA._serialized_start=2237
  _GETKINESISSOURCESCHEMA._serialized_end=2383
  _GETKAFKASOURCESCHEMA._serialized_start=2386
  _GETKAFKASOURCESCHEMA._serialized_end=2620
  _GETFEATUREVIEWSCHEMA._serialized_start=2623
  _GETFEATUREVIEWSCHEMA._serialized_end=2893
  _GETQUERYPLANINFOFORFEATUREVIEWPIPELINE._serialized_start=2896
  _GETQUERYPLANINFOFORFEATUREVIEWPIPELINE._serialized_end=3178
  _LISTHIVEDATABASES._serialized_start=3180
  _LISTHIVEDATABASES._serialized_end=3199
  _LISTHIVETABLES._serialized_start=3201
  _LISTHIVETABLES._serialized_end=3245
  _LISTHIVETABLECOLUMNS._serialized_start=3247
  _LISTHIVETABLECOLUMNS._serialized_end=3319
  _FEATUREVIEWSCHEMAREQUEST._serialized_start=3322
  _FEATUREVIEWSCHEMAREQUEST._serialized_end=3533
  _GETMULTIPLEFEATUREVIEWSCHEMASREQUEST._serialized_start=3536
  _GETMULTIPLEFEATUREVIEWSCHEMASREQUEST._serialized_end=3836
  _GETMULTIPLEFEATUREVIEWSCHEMASRESPONSE._serialized_start=3839
  _GETMULTIPLEFEATUREVIEWSCHEMASRESPONSE._serialized_end=4136
  _GETMULTIPLEFEATUREVIEWSCHEMASRESPONSE_FEATUREVIEWRESPONSESENTRY._serialized_start=4019
  _GETMULTIPLEFEATUREVIEWSCHEMASRESPONSE_FEATUREVIEWRESPONSESENTRY._serialized_end=4136
  _GETMULTIPLEDATASOURCESCHEMASREQUEST._serialized_start=4138
  _GETMULTIPLEDATASOURCESCHEMASREQUEST._serialized_end=4259
  _GETMULTIPLEDATASOURCESCHEMASRESPONSE._serialized_start=4262
  _GETMULTIPLEDATASOURCESCHEMASRESPONSE._serialized_end=4540
  _GETMULTIPLEDATASOURCESCHEMASRESPONSE_DATASOURCERESPONSESENTRY._serialized_start=4437
  _GETMULTIPLEDATASOURCESCHEMASRESPONSE_DATASOURCERESPONSESENTRY._serialized_end=4540
  _SPARKSCHEMAS._serialized_start=4543
  _SPARKSCHEMAS._serialized_end=4695
  _FEATUREVIEWSCHEMARESPONSE._serialized_start=4697
  _FEATUREVIEWSCHEMARESPONSE._serialized_end=4823
  _EXECUTEREQUEST._serialized_start=4826
  _EXECUTEREQUEST._serialized_end=6743
  _EXECUTEREQUEST_ENVVARSENTRY._serialized_start=6632
  _EXECUTEREQUEST_ENVVARSENTRY._serialized_end=6690
  _QUERYPLANINFO._serialized_start=6745
  _QUERYPLANINFO._serialized_end=6838
  _EXECUTERESULT._serialized_start=6841
  _EXECUTERESULT._serialized_end=7501
# @@protoc_insertion_point(module_scope)



import datetime
import enum
import logging
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import attrs
import pendulum
from google.protobuf import duration_pb2
from typeguard import typechecked

from tecton_core import pipeline_common
from tecton_core import specs
from tecton_core import time_utils
from tecton_core.fco_container import FcoContainer
from tecton_core.feature_view_utils import CONTINUOUS_MODE_BATCH_INTERVAL
from tecton_core.id_helper import IdHelper
from tecton_core.online_serving_index import OnlineServingIndex
from tecton_core.pipeline_common import uses_realtime_context
from tecton_core.query_consts import temp_indictor_column_name
from tecton_core.schema import Schema
from tecton_core.specs import LifetimeWindowSpec
from tecton_core.specs import RelativeTimeWindowSpec
from tecton_core.specs import TimeWindowSpec
from tecton_core.specs import create_time_window_spec_from_data_proto
from tecton_core.specs import utils
from tecton_core.time_utils import convert_timedelta_for_version
from tecton_proto.args.feature_view__client_pb2 import OfflineFeatureStoreConfig
from tecton_proto.args.pipeline__client_pb2 import DataSourceNode
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.transformation__client_pb2 import TransformationMode
from tecton_proto.common import data_source_type__client_pb2 as data_source_type_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_proto.common.framework_version__client_pb2 import FrameworkVersion as FrameworkVersionProto
from tecton_proto.data import feature_view__client_pb2 as feature_view_pb2
from tecton_proto.data.feature_view__client_pb2 import OfflineStoreParams


logger = logging.getLogger(__name__)


# Create a parallel enum class since Python proto extensions do not use an enum class.
# Keep up-to-date with FrameworkVersion from tecton_proto/common/framework_version.proto.
class FrameworkVersion(enum.Enum):
    UNSPECIFIED = FrameworkVersionProto.UNSPECIFIED
    FWV3 = FrameworkVersionProto.FWV3
    FWV5 = FrameworkVersionProto.FWV5


@attrs.define(frozen=True)
class FeatureDefinitionWrapper:
    """A container for a Feature View spec and its dependent specs, i.e. data sources, transformations, and entities."""

    fv_spec: specs.FeatureViewSpec
    fco_container: FcoContainer

    @typechecked
    def __init__(self, feature_view_spec: specs.FeatureViewSpec, fco_container: FcoContainer) -> None:
        self.__attrs_init__(  # type: ignore
            fv_spec=feature_view_spec,
            fco_container=fco_container,
        )

    @property
    def id(self) -> str:
        return self.fv_spec.id

    @property
    def name(self) -> str:
        return self.fv_spec.name

    @property
    def is_temporal_aggregate(self) -> bool:
        return (
            isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec)
            and self.fv_spec.type == specs.MaterializedFeatureViewType.TEMPORAL_AGGREGATE
        )

    @property
    def has_explicit_view_schema(self) -> bool:
        if isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            return self.fv_spec.has_explicit_view_schema
        elif isinstance(self.fv_spec, (specs.RealtimeFeatureViewSpec, specs.FeatureTableSpec)):
            # FeatureTable and ODFV are always use explicit schema.
            return True
        else:
            msg = f"FeatureViewSpec '{self.name}' does not implement `has_explicit_view_schema`"
            raise ValueError(msg)

    @property
    def is_continuous(self) -> bool:
        return isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec) and self.fv_spec.is_continuous

    @property
    def is_temporal(self) -> bool:
        return (
            isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec)
            and self.fv_spec.type == specs.MaterializedFeatureViewType.TEMPORAL
        )

    @property
    def is_feature_table(self) -> bool:
        return isinstance(self.fv_spec, specs.FeatureTableSpec)

    @property
    def is_stream(self) -> bool:
        return (
            isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec)
            and self.fv_spec.data_source_type == data_source_type_pb2.DataSourceType.STREAM_WITH_BATCH
        )

    @property
    def is_realtime(self) -> bool:
        return isinstance(self.fv_spec, specs.RealtimeFeatureViewSpec)

    @property
    def uses_realtime_context(self) -> bool:
        return uses_realtime_context(self.pipeline.root)

    @property
    def is_incremental_backfill(self) -> bool:
        return isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec) and self.fv_spec.incremental_backfills

    @property
    def get_feature_store_format_version(self) -> int:
        return self.fv_spec.feature_store_format_version

    @property
    def namespace_separator(self) -> str:
        if self.framework_version == FrameworkVersion.FWV5:
            return "__"
        else:
            return "."

    @property
    def framework_version(self) -> FrameworkVersion:
        return FrameworkVersion(self.fv_spec.metadata.framework_version)

    @property
    def time_key(self) -> Optional[str]:
        if isinstance(
            self.fv_spec,
            (specs.MaterializedFeatureViewSpec, specs.FeatureTableSpec),
        ):
            return self.fv_spec.timestamp_field
        else:
            return None

    @property
    def timestamp_key(self) -> Optional[str]:
        # TODO(jake): This property is a dupe with time_key.
        return self.time_key

    @property
    def join_keys(self) -> List[str]:
        return list(self.fv_spec.join_keys)

    @property
    def materialized_fv_spec(self) -> specs.MaterializedFeatureViewSpec:
        """
        A helper method for type safety - we often (unsafely) access properties of fv_spec that are only available for
        MaterializedFeatureViewSpecs. This helper makes it easy to have some assurance you're working with an
        instance of materialized spec in lieu of checking for `isinstance` every time.
        """
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            msg = f"Feature {self.name} of type {type(self.fv_spec)} does not have a materialized feature view spec."
            raise TypeError(msg)

        return self.fv_spec

    @property
    def join_keys_schema(self) -> Schema:
        if self.is_realtime:
            # For ODFV, we need to extract its dependent materialized FeatureViewSpec and join key override mapping to correctly build the spine schema.
            all_fv_nodes = pipeline_common.get_all_feature_view_nodes(self.pipeline)
            dependent_fv_specs = [
                self.fco_container.get_by_id_proto(node.feature_view_node.feature_view_id) for node in all_fv_nodes
            ]
            jk_overrides = [
                [
                    utils.JoinKeyMappingSpec(
                        spine_column_name=override_join_key.spine_column,
                        feature_view_column_name=override_join_key.feature_column,
                    )
                    for override_join_key in node.feature_view_node.feature_reference.override_join_keys
                ]
                for node in all_fv_nodes
            ]
            return self.fv_spec.join_key_schema(zip(dependent_fv_specs, jk_overrides))
        else:
            return self.fv_spec.join_key_schema()

    @property
    def online_serving_index(self) -> OnlineServingIndex:
        return OnlineServingIndex(list(self.fv_spec.online_serving_keys))

    @property
    def wildcard_join_key(self) -> Optional[str]:
        """
        Returns a wildcard join key column name for the feature view if it exists;
        Otherwise returns None.
        """
        online_serving_index = self.online_serving_index
        wildcard_keys = [join_key for join_key in self.join_keys if join_key not in online_serving_index.join_keys]
        return wildcard_keys[0] if wildcard_keys else None

    @property
    def aggregation_secondary_key(self) -> Optional[str]:
        if not self.is_temporal_aggregate:
            msg = f"Feature View '{self.name}' does not have aggregation_secondary_key"
            raise ValueError(msg)
        return self.fv_spec.aggregation_secondary_key

    @property
    def window_to_features_map_for_secondary_keys(self) -> Dict[TimeWindowSpec, List[str]]:
        """
        Group feature columns by windows. Each window maps to a list of columns:

        E.g. { [time_window] -> [secondary_key, secondary_key_indicator, agg_features...] }
        """
        window_to_features = {}

        for secondary_key_output in self.materialized_fv_spec.secondary_key_rollup_outputs:
            aggregation_secondary_key_and_indicator_cols = [
                self.aggregation_secondary_key,
                temp_indictor_column_name(secondary_key_output.time_window),
            ]
            features = [
                agg_feature.output_feature_name
                for agg_feature in self.trailing_time_window_aggregation().features
                if create_time_window_spec_from_data_proto(agg_feature.time_window) == secondary_key_output.time_window
            ]
            window_to_features[secondary_key_output.time_window] = (
                aggregation_secondary_key_and_indicator_cols + features
            )
        return window_to_features

    @property
    def partial_aggregate_group_by_columns(self) -> List[str]:
        """
        The columns to group by for partial aggregations. This includes the join keys of the feature view, and the
        aggregation secondary key if configured.
        """
        if not self.is_temporal_aggregate:
            msg = f"Feature View '{self.name}' does not have partial_agg_group_keys"
            raise ValueError(msg)
        return (
            [*self.join_keys, self.fv_spec.aggregation_secondary_key]
            if self.fv_spec.aggregation_secondary_key
            else self.join_keys
        )

    @property
    def has_offset_window(self) -> bool:
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            msg = f"Feature definition with type {type(self.fv_spec)} does not have offset window."
            raise TypeError(msg)
        for feature in self.fv_spec.aggregate_features:
            window_spec = create_time_window_spec_from_data_proto(feature.time_window)
            if (
                isinstance(window_spec, RelativeTimeWindowSpec)
                and window_spec.offset is not None
                and (window_spec.offset < datetime.timedelta(0))
            ):
                return True
        return False

    @property
    def has_delta_offline_store(self) -> bool:
        conf = self.offline_store_config or self.offline_store_params
        return conf.HasField("delta")

    @property
    def has_parquet_offline_store(self) -> bool:
        conf = self.offline_store_config or self.offline_store_params
        return conf.HasField("parquet")

    @property
    def offline_store_config(self) -> OfflineFeatureStoreConfig:
        if isinstance(self.fv_spec, specs.RealtimeFeatureViewSpec) or self.fv_spec.offline_store is None:
            return OfflineFeatureStoreConfig()
        return self.fv_spec.offline_store

    @property
    def online_store_data_delay_seconds(self) -> int:
        return 0 if (self.is_stream or self.is_feature_table) else self.max_source_data_delay.in_seconds()

    @property
    def materialization_enabled(self) -> bool:
        return self.fv_spec.materialization_enabled

    @property
    def writes_to_offline_store(self) -> bool:
        # Brian: I think this should actually be `return self.fv_spec.materialization_enabled and self.fv_spec.offline`
        # Otherwise this indicates we write to the offline store when we don't since materialization is disabled.
        # Similarly, it should be impossible to set offine=True on a local object?
        return self.fv_spec.offline

    @property
    def writes_to_online_store(self) -> bool:
        return self.fv_spec.online

    @property
    def view_schema(self) -> Schema:
        return self.fv_spec.view_schema

    @property
    def materialization_schema(self) -> Schema:
        return self.fv_spec.materialization_schema

    @property
    def min_scheduling_interval(self) -> Optional[pendulum.Duration]:
        if self.is_temporal_aggregate:
            return self.fv_spec.slide_interval
        elif self.is_temporal:
            return self.fv_spec.batch_schedule
        else:
            return None

    @property
    def batch_materialization_schedule(self) -> pendulum.Duration:
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            msg = f"Feature definition with type {type(self.fv_spec)} does not have a batch_materialization_schedule."
            raise TypeError(msg)

        if self.fv_spec.batch_schedule is not None:
            return self.fv_spec.batch_schedule
        elif self.fv_spec.is_continuous:
            return time_utils.proto_to_duration(CONTINUOUS_MODE_BATCH_INTERVAL)
        elif self.fv_spec.slide_interval is not None:
            return self.fv_spec.slide_interval
        else:
            msg = "Materialized feature view must have a batch_materialization_schedule."
            raise ValueError(msg)

    @property
    def offline_store_params(self) -> Optional[OfflineStoreParams]:
        return self.fv_spec.offline_store_params

    @property
    def max_source_data_delay(self) -> pendulum.Duration:
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            msg = f"Feature definition with type {type(self.fv_spec)} does not have max_source_data_delay."
            raise TypeError(msg)
        return self.fv_spec.max_source_data_delay

    @property
    def materialization_start_timestamp(self) -> Optional[pendulum.datetime]:
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            msg = f"Feature definition with type {type(self.fv_spec)} does not have a materialization_start_timestamp."
            raise TypeError(msg)

        return self.fv_spec.materialization_start_time

    @property
    def feature_start_timestamp(self) -> Optional[pendulum.datetime]:
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec) or self.fv_spec.feature_start_time is None:
            return None

        return self.fv_spec.feature_start_time

    @property
    def time_range_policy(self) -> feature_view_pb2.MaterializationTimeRangePolicy:
        if isinstance(self.fv_spec, specs.RealtimeFeatureViewSpec) or self.fv_spec.time_range_policy is None:
            msg = "No materialization time range policy set for this feature view."
            raise ValueError(msg)

        return self.fv_spec.time_range_policy

    @property
    def data_source_ids(self) -> List[str]:
        if self.pipeline is None:
            return []

        nodes = pipeline_to_ds_inputs(self.pipeline).values()
        return [IdHelper.to_string(node.virtual_data_source_id) for node in nodes]

    @property
    def data_sources(self) -> List[specs.DataSourceSpec]:
        ds_ids = self.data_source_ids
        return self.fco_container.get_by_ids(ds_ids)

    def get_data_source_with_input_name(self, input_name: str) -> specs.DataSourceSpec:
        """Get the data source spec that uses `input_name` for the feature view transformation."""
        input_name_to_ds_id = pipeline_common.get_input_name_to_ds_id_map(self.pipeline)

        if input_name not in input_name_to_ds_id:
            msg = (
                f"Feature view '{self.name}' does not have an input data source with the parameter name '{input_name}'"
            )
            raise KeyError(msg)

        return self.fco_container.get_by_id(input_name_to_ds_id[input_name])

    @property
    def get_tile_interval(self) -> pendulum.Duration:
        if self.is_temporal_aggregate:
            return self.fv_spec.slide_interval
        elif self.is_temporal:
            return self.fv_spec.batch_schedule

        msg = "Invalid invocation on unsupported FeatureView type"
        raise ValueError(msg)

    @property
    def get_batch_schedule_for_version(self) -> int:
        return time_utils.convert_timedelta_for_version(
            self.fv_spec.batch_schedule, self.get_feature_store_format_version
        )

    @property
    def get_tile_interval_for_version(self) -> int:
        if self.is_temporal_aggregate:
            return time_utils.convert_timedelta_for_version(
                self.fv_spec.slide_interval, self.get_feature_store_format_version
            )
        elif self.is_temporal:
            return time_utils.convert_timedelta_for_version(
                self.fv_spec.batch_schedule, self.get_feature_store_format_version
            )

        msg = "Invalid invocation on unsupported FeatureView type"
        raise TypeError(msg)

    @property
    def get_tile_interval_duration_for_offline_store(self) -> pendulum.Duration:
        """The tile interval to be used for offline retrieval.

        For most cases this will be the aggregation interval. For stream fvs with compaction, the offline store is always untiled, but we will still create tiles during retrieval for better performance depending on the type of aggregation windows in the fv.
        """
        if self.is_temporal_aggregate and self.is_stream and self.compaction_enabled:
            min_sawtooth_size = self.get_min_batch_sawtooth_tile_size()
            if min_sawtooth_size is None:
                return pendulum.Duration(seconds=0)
            return min_sawtooth_size
        return self.get_tile_interval

    @property
    def get_tile_interval_for_offline_store(self) -> int:
        return time_utils.convert_timedelta_for_version(
            self.get_tile_interval_duration_for_offline_store, self.get_feature_store_format_version
        )

    @property
    def get_aggregate_slide_interval_string(self) -> str:
        if not self.is_temporal_aggregate:
            msg = "Invalid invocation on unsupported FeatureView type"
            raise TypeError(msg)

        return self.fv_spec.slide_interval_string

    @property
    def aggregate_slide_interval(self) -> duration_pb2.Duration:
        if not self.is_temporal_aggregate:
            msg = "Invalid invocation on unsupported FeatureView type"
            raise TypeError(msg)

        duration = duration_pb2.Duration()
        duration.FromTimedelta(self.fv_spec.slide_interval)
        return duration

    @property
    def materialized_data_path(self) -> str:
        if isinstance(self.fv_spec, specs.RealtimeFeatureViewSpec) or self.fv_spec.materialized_data_path is None:
            msg = "No materialized data path available."
            raise ValueError(msg)

        return self.fv_spec.materialized_data_path

    @property
    def published_features_path(self) -> Optional[str]:
        """Returns the location of published features in the offline store."""
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            msg = "Invalid `published_features_path` invocation. Publish Features is not supported in Feature Tables."
            raise TypeError(msg)

        return self.fv_spec.published_features_path

    @property
    def has_lifetime_aggregate(self) -> bool:
        """Returns if FV contains a lifetime aggregate. Fails for non-agg FVs."""
        if not self.is_temporal_aggregate:
            msg = "Invalid `has_lifetime_aggregate` invocation on unsupported FeatureView type"
            raise TypeError(msg)

        for feature in self.fv_spec.aggregate_features:
            window_spec = create_time_window_spec_from_data_proto(feature.time_window)
            if isinstance(window_spec, LifetimeWindowSpec):
                return True

        return False

    @property
    def earliest_window_start(self) -> datetime.timedelta:
        """Returns the earliest window start from the window aggregates. Fails for FVs with lifetime aggregate."""
        if not self.is_temporal_aggregate:
            msg = "Invalid `earliest_window_start` invocation on unsupported FeatureView type"
            raise TypeError(msg)

        min_timedelta = datetime.timedelta()
        for feature in self.fv_spec.aggregate_features:
            window_spec = create_time_window_spec_from_data_proto(feature.time_window)
            if isinstance(window_spec, LifetimeWindowSpec):
                msg = "Invalid `earliest_window_start` invocation on FeatureView with a lifetime aggregate"
                raise ValueError(msg)
            elif isinstance(window_spec, RelativeTimeWindowSpec):
                min_timedelta = min(min_timedelta, window_spec.window_start)
            elif isinstance(window_spec, specs.TimeWindowSeriesSpec):
                min_timedelta = min(min_timedelta, window_spec.window_series_start)
        return min_timedelta

    @property
    def time_window_durations(self) -> List[datetime.timedelta]:
        if not self.is_temporal_aggregate:
            msg = "Invalid `time_window_durations` invocation on unsupported FeatureView type"
            raise TypeError(msg)
        return [
            create_time_window_spec_from_data_proto(feature.time_window).window_duration
            for feature in self.fv_spec.aggregate_features
        ]

    @property
    def transformations(self) -> List[specs.TransformationSpec]:
        if self.pipeline is None:
            return []

        transformation_ids = pipeline_to_transformation_ids(self.pipeline)
        return self.fco_container.get_by_ids(transformation_ids)

    @property
    def transformation_mode(self) -> Optional[TransformationMode.ValueType]:
        if len(self.transformations) == 0:
            return None

        # TODO: Strengthen validation here. Feature views besides pipeline mode should only have one value
        return self.transformations[0].transformation_mode

    @property
    def entities(self) -> List[specs.EntitySpec]:
        return self.fco_container.get_by_ids(self.fv_spec.entity_ids)

    def trailing_time_window_aggregation(
        self, aggregation_slide_period: Optional[pendulum.Duration] = None
    ) -> Optional[feature_view_pb2.TrailingTimeWindowAggregation]:
        if not self.is_temporal_aggregate:
            return None

        if aggregation_slide_period is not None:
            # Override the slide period if provided. This is useful during offline retrieval when we want to use a different slide period than the one used in materialization.
            aggregation_slide_period_proto = duration_pb2.Duration()
            aggregation_slide_period_proto.FromTimedelta(aggregation_slide_period)
        else:
            aggregation_slide_period_proto = self.aggregate_slide_interval

        # TODO(sanika) - trailing_time_window_aggregation fills time_window based on data proto, which is used in to_query methods.
        # This should be cleaned up when we work on ranges using offset
        return feature_view_pb2.TrailingTimeWindowAggregation(
            time_key=self.timestamp_key,
            is_continuous=self.is_continuous,
            aggregation_slide_period=aggregation_slide_period_proto,
            features=self.fv_spec.aggregate_features,
        )

    @property
    def serving_ttl(self) -> Optional[pendulum.Duration]:
        if isinstance(self.fv_spec, (specs.MaterializedFeatureViewSpec, specs.FeatureTableSpec)):
            return self.fv_spec.ttl
        else:
            return None

    @property
    def features(self) -> List[str]:
        return self.fv_spec.features

    @property
    def workspace(self) -> str:
        return self.fv_spec.workspace

    @property
    def request_context_keys(self) -> List[str]:
        rc_schema = self.request_context_schema
        if rc_schema is not None:
            return rc_schema.column_names()
        else:
            return []

    @property
    def request_context_schema(self) -> Schema:
        if self.pipeline is None:
            return Schema(schema_pb2.Schema())

        request_context = pipeline_common.find_request_context(self.pipeline.root)
        if request_context:
            return Schema(request_context.tecton_schema)
        else:
            return Schema(schema_pb2.Schema())

    # Returns the schema of the spine that can be used to query feature values. Note the actual spine user passes in
    # could contain extra columns that are not used by Tecton, and returned schema doesn't include these columns. For
    # details about how to build spine_schema for different FeatureView, see `spine_schema` method defined in
    # FeatureViewSpec and its children classes.
    @property
    def spine_schema(self) -> Schema:
        return self.join_keys_schema + self.request_context_schema

    @property
    def pipeline(self) -> Optional[Pipeline]:
        if isinstance(self.fv_spec, (specs.MaterializedFeatureViewSpec, specs.RealtimeFeatureViewSpec)):
            return self.fv_spec.pipeline
        else:
            # Feature Tables do not have pipelines.
            return None

    @property
    def manual_trigger_backfill_end_timestamp(self) -> Optional[pendulum.DateTime]:
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            msg = f"Feature definition with type {type(self.fv_spec)} does not have a manual_trigger_backfill_end_time."
            raise TypeError(msg)

        return self.fv_spec.manual_trigger_backfill_end_time

    def earliest_anchor_time_from_window_start(self, window_start: Optional[datetime.timedelta]) -> Optional[int]:
        if window_start is None:
            return None
        if self.is_continuous and not self.compaction_enabled:
            # We do + 1 since RangeBetween is inclusive, and we do not want to include the last row of the
            # previous tile. See https://github.com/tecton-ai/tecton/pull/1110
            tile_interval = 1
        else:
            tile_interval = self.get_tile_interval_for_offline_store
        return (
            convert_timedelta_for_version(window_start, version=self.get_feature_store_format_version) + tile_interval
        )

    def time_range_for_relative_time_window(self, time_window: RelativeTimeWindowSpec) -> Tuple[int, int]:
        start = self.earliest_anchor_time_from_window_start(time_window.window_start)
        end = convert_timedelta_for_version(time_window.window_end, self.get_feature_store_format_version)
        assert start <= end
        return start, end

    @property
    def compaction_enabled(self) -> bool:
        """Whether a feature view has `compaction_enabled` set to True."""
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            return False
        return self.fv_spec.compaction_enabled

    @property
    def stream_tiling_enabled(self) -> bool:
        """Whether a stream feature view has stream_tiling_enabled set to True.

        stream_tiling_enabled is only applicable to stream feature views with compaction_enabled=True."""
        if not isinstance(self.fv_spec, specs.MaterializedFeatureViewSpec):
            return False
        return self.fv_spec.stream_tiling_enabled

    def get_min_batch_sawtooth_tile_size(self) -> Optional[pendulum.Duration]:
        """Returns the smallest batch sawtooth tile size for each aggregation in a fv.

        This is only non-null for stream fvs that use compaction."""
        if not self.is_temporal_aggregate:
            return None

        sawtooth_tiles = self.get_batch_sawtooth_tile_sizes()
        return min(sawtooth_tiles) if sawtooth_tiles else None

    def get_batch_sawtooth_tile_sizes(self) -> List[pendulum.Duration]:
        """Returns the batch sawtooth tile size for each aggregation in a fv.

        Only applicable for stream fvs that use compaction."""
        if not self.is_temporal_aggregate:
            return []

        return [
            time_utils.proto_to_duration(agg_feature.batch_sawtooth_tile_size)
            for agg_feature in self.fv_spec.aggregate_features
            if agg_feature.HasField("batch_sawtooth_tile_size")
        ]


def pipeline_to_ds_inputs(pipeline: Pipeline) -> Dict[str, DataSourceNode]:
    ds_nodes: Dict[str, DataSourceNode] = {}

    def _recurse_pipeline_to_ds_nodes(pipeline_node: PipelineNode, ds_nodes_: Dict[str, DataSourceNode]) -> None:
        if pipeline_node.HasField("data_source_node"):
            ds_nodes_[pipeline_node.data_source_node.input_name] = pipeline_node.data_source_node
        elif pipeline_node.HasField("transformation_node"):
            inputs = pipeline_node.transformation_node.inputs
            for input_ in inputs:
                _recurse_pipeline_to_ds_nodes(input_.node, ds_nodes_)

    _recurse_pipeline_to_ds_nodes(pipeline.root, ds_nodes)

    return ds_nodes


def pipeline_to_transformation_ids(pipeline: Pipeline) -> List[str]:
    id_list: List[str] = []

    def _recurse_pipeline_to_transformation_ids(node: PipelineNode, id_list: List[str]) -> List[str]:
        if node.HasField("transformation_node"):
            id_list.append(IdHelper.to_string(node.transformation_node.transformation_id))
            for input in node.transformation_node.inputs:
                _recurse_pipeline_to_transformation_ids(input.node, id_list)
        return id_list

    _recurse_pipeline_to_transformation_ids(pipeline.root, id_list)
    return id_list



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/principal_group__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.auditlog import metadata__client_pb2 as tecton__proto_dot_auditlog_dot_metadata____client__pb2
from tecton_proto.auth import principal__client_pb2 as tecton__proto_dot_auth_dot_principal____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/tecton_proto/data/principal_group__client.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a,tecton_proto/auditlog/metadata__client.proto\x1a)tecton_proto/auth/principal__client.proto\"\x86\x03\n\x0ePrincipalGroup\x12\x17\n\x02id\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x02id\x12\x1b\n\x04name\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x04name\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\x12\x33\n\x11idp_mapping_names\x18\x04 \x03(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x0fidpMappingNames\x12\x34\n\x16is_membership_editable\x18\x05 \x01(\x08R\x14isMembershipEditable\x12\x39\n\ncreated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12;\n\ncreated_by\x18\x07 \x01(\x0b\x32\x1c.tecton_proto.auth.PrincipalR\tcreatedBy\x12\x39\n\nupdated_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAtB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.principal_group__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _PRINCIPALGROUP.fields_by_name['id']._options = None
  _PRINCIPALGROUP.fields_by_name['id']._serialized_options = b'\352\231\275F\002\010\002'
  _PRINCIPALGROUP.fields_by_name['name']._options = None
  _PRINCIPALGROUP.fields_by_name['name']._serialized_options = b'\352\231\275F\002\010\002'
  _PRINCIPALGROUP.fields_by_name['idp_mapping_names']._options = None
  _PRINCIPALGROUP.fields_by_name['idp_mapping_names']._serialized_options = b'\352\231\275F\002\010\002'
  _PRINCIPALGROUP._serialized_start=193
  _PRINCIPALGROUP._serialized_end=583
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/user_defined_function__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5tecton_proto/args/user_defined_function__client.proto\x12\x11tecton_proto.args\x1a,tecton_proto/args/diff_options__client.proto\"J\n\x13UserDefinedFunction\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x19\n\x04\x62ody\x18\x03 \x01(\tB\x05\x92M\x02\x18\x02R\x04\x62odyJ\x04\x08\x02\x10\x03\x42;\n\x0f\x63om.tecton.argsP\x01Z&github.com/tecton-ai/tecton_proto/args')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.user_defined_function__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001Z&github.com/tecton-ai/tecton_proto/args'
  _USERDEFINEDFUNCTION.fields_by_name['body']._options = None
  _USERDEFINEDFUNCTION.fields_by_name['body']._serialized_options = b'\222M\002\030\002'
  _USERDEFINEDFUNCTION._serialized_start=122
  _USERDEFINEDFUNCTION._serialized_end=196
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import warnings
from collections import Counter

from pyspark.rdd import _load_from_socket
from pyspark.sql.pandas.serializers import ArrowCollectSerializer
from pyspark.sql.types import IntegralType
from pyspark.sql.types import ByteType, ShortType, IntegerType, LongType, FloatType, \
    DoubleType, BooleanType, MapType, TimestampType, StructType, DataType
from pyspark.traceback_utils import SCCallSiteSync


class PandasConversionMixin(object):
    """
    Min-in for the conversion from Spark to pandas. Currently, only :class:`DataFrame`
    can use this class.
    """

    def toPandas(self):
        """
        Returns the contents of this :class:`DataFrame` as Pandas ``pandas.DataFrame``.

        This is only available if Pandas is installed and available.

        .. versionadded:: 1.3.0

        Notes
        -----
        This method should only be used if the resulting Pandas's :class:`DataFrame` is
        expected to be small, as all the data is loaded into the driver's memory.

        Usage with spark.sql.execution.arrow.pyspark.enabled=True is experimental.

        Examples
        --------
        >>> df.toPandas()  # doctest: +SKIP
           age   name
        0    2  Alice
        1    5    Bob
        """
        from tecton_spark.type_annotations import PySparkDataFrame

        assert isinstance(self, PySparkDataFrame)

        from pyspark.sql.pandas.utils import require_minimum_pandas_version
        require_minimum_pandas_version()

        import numpy as np
        import pandas as pd

        timezone = self.sql_ctx._conf.sessionLocalTimeZone()

        if self.sql_ctx._conf.arrowPySparkEnabled():
            use_arrow = True
            try:
                from pyspark.sql.pandas.types import to_arrow_schema
                from pyspark.sql.pandas.utils import require_minimum_pyarrow_version

                require_minimum_pyarrow_version()
                to_arrow_schema(self.schema)
            except Exception as e:

                if self.sql_ctx._conf.arrowPySparkFallbackEnabled():
                    msg = (
                        "toPandas attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true; however, "
                        "failed by the reason below:\n  %s\n"
                        "Attempting non-optimization as "
                        "'spark.sql.execution.arrow.pyspark.fallback.enabled' is set to "
                        "true." % str(e))
                    warnings.warn(msg)
                    use_arrow = False
                else:
                    msg = (
                        "toPandas attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                        "reached the error below and will not continue because automatic fallback "
                        "with 'spark.sql.execution.arrow.pyspark.fallback.enabled' has been set to "
                        "false.\n  %s" % str(e))
                    warnings.warn(msg)
                    raise

            # Try to use Arrow optimization when the schema is supported and the required version
            # of PyArrow is found, if 'spark.sql.execution.arrow.pyspark.enabled' is enabled.
            if use_arrow:
                try:
                    from pyspark.sql.pandas.types import _check_series_localize_timestamps, \
                        _convert_map_items_to_dict
                    import pyarrow
                    # Rename columns to avoid duplicated column names.
                    tmp_column_names = ['col_{}'.format(i) for i in range(len(self.columns))]
                    batches = self.toDF(*tmp_column_names)._collect_as_arrow()
                    if len(batches) > 0:
                        table = pyarrow.Table.from_batches(batches)
                        # Pandas DataFrame created from PyArrow uses datetime64[ns] for date type
                        # values, but we should use datetime.date to match the behavior with when
                        # Arrow optimization is disabled.
                        pdf = table.to_pandas(date_as_object=True)
                        # Rename back to the original column names.
                        pdf.columns = self.columns
                        for field in self.schema:
                            if isinstance(field.dataType, TimestampType):
                                pdf[field.name] = \
                                    _check_series_localize_timestamps(pdf[field.name], timezone)
                            elif isinstance(field.dataType, MapType):
                                pdf[field.name] = \
                                    _convert_map_items_to_dict(pdf[field.name])
                        return pdf
                    else:
                        return pd.DataFrame.from_records([], columns=self.columns)
                except Exception as e:
                    # We might have to allow fallback here as well but multiple Spark jobs can
                    # be executed. So, simply fail in this case for now.
                    msg = (
                        "toPandas attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                        "reached the error below and can not continue. Note that "
                        "'spark.sql.execution.arrow.pyspark.fallback.enabled' does not have an "
                        "effect on failures in the middle of "
                        "computation.\n  %s" % str(e))
                    warnings.warn(msg)
                    raise

        # Below is toPandas without Arrow optimization.
        pdf = pd.DataFrame.from_records(self.collect(), columns=self.columns)
        column_counter = Counter(self.columns)

        dtype = [None] * len(self.schema)
        for fieldIdx, field in enumerate(self.schema):
            # For duplicate column name, we use `iloc` to access it.
            if column_counter[field.name] > 1:
                pandas_col = pdf.iloc[:, fieldIdx]
            else:
                pandas_col = pdf[field.name]

            pandas_type = PandasConversionMixin._to_corrected_pandas_type(field.dataType)
            # SPARK-21766: if an integer field is nullable and has null values, it can be
            # inferred by pandas as float column. Once we convert the column with NaN back
            # to integer type e.g., np.int16, we will hit exception. So we use the inferred
            # float type, not the corrected type from the schema in this case.
            if pandas_type is not None and \
                not(isinstance(field.dataType, IntegralType) and field.nullable and
                    pandas_col.isnull().any()):
                dtype[fieldIdx] = pandas_type
            # Ensure we fall back to nullable numpy types, even when whole column is null:
            if isinstance(field.dataType, IntegralType) and pandas_col.isnull().any():
                dtype[fieldIdx] = np.float64
            if isinstance(field.dataType, BooleanType) and pandas_col.isnull().any():
                dtype[fieldIdx] = np.object

        df = pd.DataFrame()
        for index, t in enumerate(dtype):
            column_name = self.schema[index].name

            # For duplicate column name, we use `iloc` to access it.
            if column_counter[column_name] > 1:
                series = pdf.iloc[:, index]
            else:
                series = pdf[column_name]

            if t is not None:
                series = series.astype(t, copy=False)

            # `insert` API makes copy of data, we only do it for Series of duplicate column names.
            # `pdf.iloc[:, index] = pdf.iloc[:, index]...` doesn't always work because `iloc` could
            # return a view or a copy depending by context.
            if column_counter[column_name] > 1:
                df.insert(index, column_name, series, allow_duplicates=True)
            else:
                df[column_name] = series

        pdf = df

        if timezone is None:
            return pdf
        else:
            from pyspark.sql.pandas.types import _check_series_convert_timestamps_local_tz
            for field in self.schema:
                # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
                if isinstance(field.dataType, TimestampType):
                    pdf[field.name] = \
                        _check_series_convert_timestamps_local_tz(pdf[field.name], timezone)
            return pdf

    @staticmethod
    def _to_corrected_pandas_type(dt):
        """
        When converting Spark SQL records to Pandas :class:`DataFrame`, the inferred data type
        may be wrong. This method gets the corrected data type for Pandas if that type may be
        inferred incorrectly.
        """
        import numpy as np
        if type(dt) == ByteType:
            return np.int8
        elif type(dt) == ShortType:
            return np.int16
        elif type(dt) == IntegerType:
            return np.int32
        elif type(dt) == LongType:
            return np.int64
        elif type(dt) == FloatType:
            return np.float32
        elif type(dt) == DoubleType:
            return np.float64
        elif type(dt) == BooleanType:
            return np.bool
        elif type(dt) == TimestampType:
            return np.datetime64
        else:
            return None

    def _collect_as_arrow(self):
        """
        Returns all records as a list of ArrowRecordBatches, pyarrow must be installed
        and available on driver and worker Python environments.
        This is an experimental feature.
        """
        from tecton_spark.type_annotations import PySparkDataFrame

        assert isinstance(self, PySparkDataFrame)

        with SCCallSiteSync(self._sc):
            port, auth_secret, jsocket_auth_server = self._jdf.collectAsArrowToPython()

        # Collect list of un-ordered batches where last element is a list of correct order indices
        try:
            results = list(_load_from_socket((port, auth_secret), ArrowCollectSerializer()))
        finally:
            # Join serving thread and raise any exceptions from collectAsArrowToPython
            jsocket_auth_server.getResult()

        # Separate RecordBatches from batch order indices in results
        batches = results[:-1]
        batch_order = results[-1]

        # Re-order the batch list using the correct order
        return [batches[i] for i in batch_order]


class SparkConversionMixin(object):
    """
    Min-in for the conversion from pandas to Spark. Currently, only :class:`SparkSession`
    can use this class.
    """
    def createDataFrame(self, data, schema=None, samplingRatio=None, verifySchema=True):
        from tecton_spark.type_annotations import PySparkSession

        assert isinstance(self, PySparkSession)

        from pyspark.sql.pandas.utils import require_minimum_pandas_version
        require_minimum_pandas_version()

        timezone = self._wrapped._conf.sessionLocalTimeZone()

        # If no schema supplied by user then get the names of columns only
        if schema is None:
            schema = [str(x) if not isinstance(x, str) else
                      (x.encode('utf-8') if not isinstance(x, str) else x)
                      for x in data.columns]

        if self._wrapped._conf.arrowPySparkEnabled() and len(data) > 0:
            try:
                return self._create_from_pandas_with_arrow(data, schema, timezone)
            except Exception as e:
                if self._wrapped._conf.arrowPySparkFallbackEnabled():
                    msg = (
                        "createDataFrame attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true; however, "
                        "failed by the reason below:\n  %s\n"
                        "Attempting non-optimization as "
                        "'spark.sql.execution.arrow.pyspark.fallback.enabled' is set to "
                        "true." % str(e))
                    warnings.warn(msg)
                else:
                    msg = (
                        "createDataFrame attempted Arrow optimization because "
                        "'spark.sql.execution.arrow.pyspark.enabled' is set to true, but has "
                        "reached the error below and will not continue because automatic "
                        "fallback with 'spark.sql.execution.arrow.pyspark.fallback.enabled' "
                        "has been set to false.\n  %s" % str(e))
                    warnings.warn(msg)
                    raise
        data = self._convert_from_pandas(data, schema, timezone)
        return self._create_dataframe(data, schema, samplingRatio, verifySchema)

    def _convert_from_pandas(self, pdf, schema, timezone):
        """
         Convert a pandas.DataFrame to list of records that can be used to make a DataFrame

         Returns
         -------
         list
             list of records
        """
        from pyspark.sql import SparkSession

        assert isinstance(self, SparkSession)

        if timezone is not None:
            from pyspark.sql.pandas.types import _check_series_convert_timestamps_tz_local
            copied = False
            if isinstance(schema, StructType):
                for field in schema:
                    # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
                    if isinstance(field.dataType, TimestampType):
                        s = _check_series_convert_timestamps_tz_local(pdf[field.name], timezone)
                        if s is not pdf[field.name]:
                            if not copied:
                                # Copy once if the series is modified to prevent the original
                                # Pandas DataFrame from being updated
                                pdf = pdf.copy()
                                copied = True
                            pdf[field.name] = s
            else:
                for column, series in pdf.iteritems():
                    s = _check_series_convert_timestamps_tz_local(series, timezone)
                    if s is not series:
                        if not copied:
                            # Copy once if the series is modified to prevent the original
                            # Pandas DataFrame from being updated
                            pdf = pdf.copy()
                            copied = True
                        pdf[column] = s

        # Convert pandas.DataFrame to list of numpy records
        np_records = pdf.to_records(index=False)

        # Check if any columns need to be fixed for Spark to infer properly
        if len(np_records) > 0:
            record_dtype = self._get_numpy_record_dtype(np_records[0])
            if record_dtype is not None:
                return [r.astype(record_dtype).tolist() for r in np_records]

        # Convert list of numpy records to python lists
        return [r.tolist() for r in np_records]

    def _get_numpy_record_dtype(self, rec):
        """
        Used when converting a pandas.DataFrame to Spark using to_records(), this will correct
        the dtypes of fields in a record so they can be properly loaded into Spark.

        Parameters
        ----------
        rec : numpy.record
            a numpy record to check field dtypes

        Returns
        -------
        numpy.dtype
            corrected dtype for a numpy.record or None if no correction needed
        """
        import numpy as np
        cur_dtypes = rec.dtype
        col_names = cur_dtypes.names
        record_type_list = []
        has_rec_fix = False
        for i in range(len(cur_dtypes)):
            curr_type = cur_dtypes[i]
            # If type is a datetime64 timestamp, convert to microseconds
            # NOTE: if dtype is datetime[ns] then np.record.tolist() will output values as longs,
            # conversion from [us] or lower will lead to py datetime objects, see SPARK-22417
            if curr_type == np.dtype('datetime64[ns]'):
                curr_type = 'datetime64[us]'
                has_rec_fix = True
            record_type_list.append((str(col_names[i]), curr_type))
        return np.dtype(record_type_list) if has_rec_fix else None

    def _create_from_pandas_with_arrow(self, pdf, schema, timezone):
        """
        Create a DataFrame from a given pandas.DataFrame by slicing it into partitions, converting
        to Arrow data, then sending to the JVM to parallelize. If a schema is passed in, the
        data types will be used to coerce the data in Pandas to Arrow conversion.
        """
        from tecton_spark.type_annotations import PySparkSession
        from pyspark.sql.dataframe import DataFrame

        assert isinstance(self, PySparkSession)

        from pyspark.sql.pandas.serializers import ArrowStreamPandasSerializer
        from pyspark.sql.types import TimestampType
        from pyspark.sql.pandas.types import from_arrow_type, to_arrow_type
        from pyspark.sql.pandas.utils import require_minimum_pandas_version, \
            require_minimum_pyarrow_version

        require_minimum_pandas_version()
        require_minimum_pyarrow_version()

        from pandas.api.types import is_datetime64_dtype, is_datetime64tz_dtype
        import pyarrow as pa

        # Create the Spark schema from list of names passed in with Arrow types
        if isinstance(schema, (list, tuple)):
            arrow_schema = pa.Schema.from_pandas(pdf, preserve_index=False)
            struct = StructType()
            for name, field in zip(schema, arrow_schema):
                struct.add(name, from_arrow_type(field.type), nullable=field.nullable)
            schema = struct

        # Determine arrow types to coerce data when creating batches
        if isinstance(schema, StructType):
            arrow_types = [to_arrow_type(f.dataType) for f in schema.fields]
        elif isinstance(schema, DataType):
            raise ValueError("Single data type %s is not supported with Arrow" % str(schema))
        else:
            # Any timestamps must be coerced to be compatible with Spark
            arrow_types = [to_arrow_type(TimestampType())
                           if is_datetime64_dtype(t) or is_datetime64tz_dtype(t) else None
                           for t in pdf.dtypes]

        # Slice the DataFrame to be batched
        step = -(-len(pdf) // self.sparkContext.defaultParallelism)  # round int up
        pdf_slices = (pdf.iloc[start:start + step] for start in range(0, len(pdf), step))

        # Create list of Arrow (columns, type) for serializer dump_stream
        arrow_data = [[(c, t) for (_, c), t in zip(pdf_slice.iteritems(), arrow_types)]
                      for pdf_slice in pdf_slices]

        jsqlContext = self._wrapped._jsqlContext

        safecheck = self._wrapped._conf.arrowSafeTypeConversion()
        col_by_name = True  # col by name only applies to StructType columns, can't happen here
        ser = ArrowStreamPandasSerializer(timezone, safecheck, col_by_name)

        def reader_func(temp_filename):
            return self._jvm.PythonSQLUtils.readArrowStreamFromFile(jsqlContext, temp_filename)

        def create_RDD_server():
            return self._jvm.ArrowRDDServer(jsqlContext)

        # Create Spark DataFrame from Arrow stream file, using one batch per partition
        jrdd = self._sc._serialize_to_jvm(arrow_data, ser, reader_func, create_RDD_server)
        jdf = self._jvm.PythonSQLUtils.toDataFrame(jrdd, schema.json(), jsqlContext)
        df = DataFrame(jdf, self._wrapped)
        df._schema = schema
        return df


def _test():
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.sql.pandas.conversion
    globs = pyspark.sql.pandas.conversion.__dict__.copy()
    spark = SparkSession.builder\
        .master("local[4]")\
        .appName("sql.pandas.conversion tests")\
        .getOrCreate()
    globs['spark'] = spark
    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.pandas.conversion, globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE | doctest.REPORT_NDIFF)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



from __future__ import annotations

import datetime
import json
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import attrs
import pandas
from pyspark.sql import dataframe as pyspark_dataframe
from pyspark.sql import streaming as pyspark_streaming
from typeguard import typechecked

from tecton import types
from tecton._internals import display
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals import querytree_api
from tecton._internals import sdk_decorators
from tecton._internals import snowflake_api
from tecton._internals import spark_api
from tecton._internals import type_utils
from tecton._internals import validations_api
from tecton._internals.ingestion import IngestionClient
from tecton.framework import base_tecton_object
from tecton.framework import configs
from tecton.framework import data_frame
from tecton_core import conf
from tecton_core import id_helper
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.compute_mode import offline_retrieval_compute_mode
from tecton_core.repo_file_handler import construct_fco_source_info
from tecton_proto.args import basic_info__client_pb2 as basic_info_pb2
from tecton_proto.args import fco_args__client_pb2 as fco_args_pb2
from tecton_proto.args import virtual_data_source__client_pb2 as virtual_data_source__args_pb2
from tecton_proto.common import data_source_type__client_pb2 as data_source_type_pb2
from tecton_proto.common import fco_locator__client_pb2 as fco_locator_pb2
from tecton_proto.common import framework_version__client_pb2 as framework_version_pb2
from tecton_proto.common import schema_container__client_pb2 as schema_container_pb2
from tecton_proto.common import spark_schema__client_pb2 as spark_schema_pb2
from tecton_proto.common.spark_schema__client_pb2 import SparkSchema
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2
from tecton_spark import spark_schema_wrapper


BatchConfigType = Union[
    configs.FileConfig,
    configs.HiveConfig,
    configs.RedshiftConfig,
    configs.SnowflakeConfig,
    configs.SparkBatchConfig,
    configs.UnityConfig,
    configs.PandasBatchConfig,
    configs.BigQueryConfig,
]

StreamConfigType = Union[configs.KinesisConfig, configs.KafkaConfig, configs.SparkStreamConfig, configs.PushConfig]

logger = logging.getLogger(__name__)


@attrs.define(eq=False)
class DataSource(base_tecton_object.BaseTectonObject):
    """Base class for Data Source classes."""

    # A data source spec, i.e. a dataclass representation of the Tecton object that is used in most functional use
    # cases, e.g. constructing queries. Set only after the object has been validated. Remote objects, i.e. applied
    # objects fetched from the backend, are assumed valid.
    _spec: Optional[specs.DataSourceSpec] = attrs.field(repr=False)

    # A Tecton "args" proto. Only set if this object was defined locally, i.e. this object was not applied and
    # fetched from the Tecton backend.
    _args: Optional[virtual_data_source__args_pb2.VirtualDataSourceArgs] = attrs.field(
        repr=False, on_setattr=attrs.setters.frozen
    )

    # A supplement to the _args proto that is needed to create the Data Source spec.
    _args_supplement: Optional[specs.DataSourceSpecArgsSupplement] = attrs.field(
        repr=False, on_setattr=attrs.setters.frozen
    )

    @sdk_decorators.assert_local_object
    def _build_args(self) -> fco_args_pb2.FcoArgs:
        return fco_args_pb2.FcoArgs(virtual_data_source=self._args)

    def _build_fco_validation_args(self) -> validator_pb2.FcoValidationArgs:
        if self.info._is_local_object:
            return validator_pb2.FcoValidationArgs(
                virtual_data_source=validator_pb2.VirtualDataSourceValidationArgs(
                    args=self._args,
                    batch_schema=self._args_supplement.batch_schema,
                    stream_schema=self._args_supplement.stream_schema,
                )
            )
        else:
            return self._spec.validation_args

    def _validate(self, indentation_level: int = 0) -> None:
        validations_api.run_backend_validation_and_assert_valid(
            self,
            validator_pb2.ValidationRequest(validation_args=[self._build_fco_validation_args()]),
            indentation_level,
        )

    def _populate_spec(self):
        # TODO(TEC-19869): Remove dependencies on this field, either switching to lazy-evaluation
        # if relevant or removing. We no longer eagerly derive the schema for datasources, so it
        # is not usable.
        self._args_supplement.batch_schema = spark_schema_pb2.SparkSchema()
        self._args_supplement.stream_schema = spark_schema_pb2.SparkSchema()
        self._spec = specs.DataSourceSpec.from_args_proto(self._args, self._args_supplement)

    @sdk_decorators.assert_local_object
    def _create_unvalidated_spec(
        self, mock_data: Union[pandas.DataFrame, pyspark_dataframe.DataFrame]
    ) -> specs.DataSourceSpec:
        """Create an unvalidated spec. Used for user unit testing, where backend validation is unavailable."""
        if isinstance(mock_data, pyspark_dataframe.DataFrame):
            schema = spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(mock_data.schema)
        elif isinstance(mock_data, pandas.DataFrame):
            # Using a empty schema for Rift, since Rift does not use the schema for batch sources.
            schema = SparkSchema()
        else:
            msg = f"Unexpected mock source type {type(mock_data)}"
            raise TypeError(msg)
        # Use the mock schema as both the batch and stream schema because StreamSource specs expect a non-nil stream
        # schema.
        supplement = attrs.evolve(self._args_supplement, batch_schema=schema, stream_schema=schema)
        return specs.DataSourceSpec.from_args_proto(self._args, supplement)

    @sdk_decorators.sdk_public_method
    @sdk_decorators.assert_remote_object
    def summary(self) -> display.Displayable:
        """Displays a human readable summary of this Data Source."""
        request = metadata_service_pb2.GetVirtualDataSourceSummaryRequest(
            fco_locator=fco_locator_pb2.FcoLocator(id=self._spec.id_proto, workspace=self._spec.workspace)
        )
        response = metadata_service.instance().GetVirtualDataSourceSummary(request)
        return display.Displayable.from_fco_summary(response.fco_summary)

    @sdk_decorators.sdk_public_method
    def get_dataframe(
        self,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        *,
        apply_translator: bool = True,
        compute_mode: Optional[Union[ComputeMode, str]] = None,
    ) -> data_frame.TectonDataFrame:
        """Returns the data in this Data Source as a Tecton DataFrame.

        :param start_time: The interval start time from when we want to retrieve source data.
            If no timezone is specified, will default to using UTC.
            Can only be defined if ``apply_translator`` is True.
        :param end_time: The interval end time until when we want to retrieve source data.
            If no timezone is specified, will default to using UTC.
            Can only be defined if ``apply_translator`` is True.
        :param apply_translator: If True, the transformation specified by ``post_processor``
            will be applied to the dataframe for the data source. ``apply_translator`` is not applicable
            to batch sources configured with ``spark_batch_config`` because it does not have a
            ``post_processor``.
        :param compute_mode: Compute mode to use to produce the data frame.

        :return: A Tecton DataFrame containing the data source's raw or translated source data.

        :raises TectonValidationError: If ``apply_translator`` is False, but ``start_time`` or
            ``end_time`` filters are passed in.
        """
        compute_mode = offline_retrieval_compute_mode(compute_mode)
        _apply_translator = apply_translator
        if self._spec.type == data_source_type_pb2.DataSourceType.PUSH_NO_BATCH:
            if self._args is not None:
                # Object defined locally, and we can't really call get_dataframe on that.
                raise errors.DATA_SOURCE_HAS_NO_BATCH_CONFIG(self.name)
            else:
                _apply_translator = False

        if compute_mode == ComputeMode.SNOWFLAKE and conf.get_bool("USE_DEPRECATED_SNOWFLAKE_RETRIEVAL"):
            return snowflake_api.get_dataframe_for_data_source(self._spec, start_time, end_time)
        else:
            return querytree_api.get_dataframe_for_data_source(
                compute_mode.default_dialect(), compute_mode, self._spec, start_time, end_time, _apply_translator
            )

    @property
    def _data_source_type(self) -> data_source_type_pb2.DataSourceType.ValueType:
        return self._spec.type

    @property
    def data_delay(self) -> Optional[datetime.timedelta]:
        """Returns the duration that materialization jobs wait after the ``batch_schedule`` before starting, typically to ensure that all data has landed."""
        return self._spec.batch_source.data_delay if self._spec.batch_source is not None else None


@attrs.define(eq=False)
class BatchSource(DataSource):
    """A Tecton BatchSource, used to read batch data into Tecton for use in a BatchFeatureView.

    Example of a BatchSource declaration:

    .. code-block:: python

        # Declare a BatchSource with a HiveConfig instance as its batch_config parameter.
        # Refer to the "Configs Classes and Helpers" section for other batch_config types.
        from tecton import HiveConfig, BatchSource

        credit_scores_batch = BatchSource(
            name='credit_scores_batch',
            batch_config=HiveConfig(
                database='demo_fraud',
                table='credit_scores',
                timestamp_field='timestamp'
            ),
            owner='matt@tecton.ai',
            tags={'release': 'production'}
        )
    """

    @typechecked
    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
        prevent_destroy: bool = False,
        batch_config: BatchConfigType,
        options: Optional[Dict[str, str]] = None,
    ):
        """Creates a new BatchSource.

        :param name: A unique name of the DataSource.
        :param description: A human-readable description.
        :param tags: Tags associated with this Tecton Data Source (key-value pairs of arbitrary metadata).
        :param owner: Owner name (typically the email of the primary maintainer).
        :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
            destructive update) during tecton plan/apply. To remove or update this object, `prevent_destroy` must be
            set to False via the same tecton apply or a separate tecton apply. `prevent_destroy` can be used to prevent
            accidental changes such as inadvertantly deleting a Feature Service used in production or recreating a Feature
            View that triggers expensive rematerialization jobs. `prevent_destroy` also blocks changes to dependent Tecton
            objects that would trigger a recreate of the tagged object, e.g. if `prevent_destroy` is set on a Feature Service,
            that will also prevent deletions or re-creates of Feature Views used in that service. `prevent_destroy` is
            only enforced in live (i.e. non-dev) workspaces.
        :param batch_config: BatchConfig object containing the configuration of the Batch Data Source to be included
            in this Data Source.
        :param options: Additional options to configure the Source. Used for advanced use cases and beta features.
        """
        ds_args = virtual_data_source__args_pb2.VirtualDataSourceArgs(
            virtual_data_source_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            version=framework_version_pb2.FrameworkVersion.FWV5,
            type=data_source_type_pb2.DataSourceType.BATCH,
            prevent_destroy=prevent_destroy,
            options=options,
        )
        batch_config._merge_batch_args(ds_args)

        info = base_tecton_object.TectonObjectInfo.from_args_proto(ds_args.info, ds_args.virtual_data_source_id)
        source_info = construct_fco_source_info(ds_args.virtual_data_source_id)

        self.__attrs_init__(
            info=info,
            spec=None,
            args=ds_args,
            source_info=source_info,
            args_supplement=_build_args_supplement(batch_config, None),
        )
        if not conf.get_bool("TECTON_SKIP_OBJECT_VALIDATION"):
            self._validate()
        self._populate_spec()
        base_tecton_object._register_local_object(self)

    @classmethod
    @typechecked
    def _from_spec(cls, spec: specs.DataSourceSpec) -> "BatchSource":
        """Create a BatchSource from directly from a spec. Specs are assumed valid and will not be re-validated."""
        info = base_tecton_object.TectonObjectInfo.from_spec(spec)
        obj = cls.__new__(cls)  # Instantiate the object. Does not call init.
        obj.__attrs_init__(
            info=info,
            spec=spec,
            args=None,
            source_info=None,
            args_supplement=None,
        )
        return obj


@attrs.define(eq=False)
class StreamSource(DataSource):
    """A Tecton StreamSource, used to unify stream and batch data into Tecton for use in a StreamFeatureView.

    Example of a StreamSource declaration:

    .. code-block:: python

     import pyspark
        from tecton import KinesisConfig, HiveConfig, StreamSource
        from datetime import timedelta


        # Define our deserialization raw stream translator
        def raw_data_deserialization(df:pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
            from pyspark.sql.functions import col, from_json, from_utc_timestamp
            from pyspark.sql.types import StructType, StringType

            payload_schema = (
              StructType()
                    .add('amount', StringType(), False)
                    .add('isFraud', StringType(), False)
                    .add('timestamp', StringType(), False)
            )
            return (
                df.selectExpr('cast (data as STRING) jsonData')
                .select(from_json('jsonData', payload_schema).alias('payload'))
                .select(
                    col('payload.amount').cast('long').alias('amount'),
                    col('payload.isFraud').cast('long').alias('isFraud'),
                    from_utc_timestamp('payload.timestamp', 'UTC').alias('timestamp')
                )
            )

        # Declare a StreamSource with both a batch_config and a stream_config as parameters
        # See the API documentation for both BatchConfig and StreamConfig
        transactions_stream = StreamSource(
            name='transactions_stream',
            stream_config=KinesisConfig(
                stream_name='transaction_events',
                region='us-west-2',
                initial_stream_position='latest',
                watermark_delay_threshold=timedelta(minutes=30),
                timestamp_field='timestamp',
                post_processor=raw_data_deserialization, # deserialization function defined above
                options={'roleArn': 'arn:aws:iam::472542229217:role/demo-cross-account-kinesis-ro'}
            ),
            batch_config=HiveConfig(
                database='demo_fraud',
                table='transactions',
                timestamp_field='timestamp',
            ),
            owner='user@tecton.ai',
            tags={'release': 'staging'}
        )
    """

    @typechecked
    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
        prevent_destroy: bool = False,
        batch_config: Optional[BatchConfigType] = None,
        stream_config: StreamConfigType,
        options: Optional[Dict[str, str]] = None,
        schema: Optional[List[types.Field]] = None,
    ):
        """Creates a new StreamSource.

        :param name: A unique name of the DataSource.
        :param description: A human-readable description.
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param owner: Owner name (typically the email of the primary maintainer).
        :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
            destructive update) during tecton plan/apply. To remove or update this object, `prevent_destroy` must be
            set to False via the same tecton apply or a separate tecton apply. `prevent_destroy` can be used to prevent accidental changes
            such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
            triggers expensive rematerialization jobs. `prevent_destroy` also blocks changes to dependent Tecton objects
            that would trigger a recreate of the tagged object, e.g. if `prevent_destroy` is set on a Feature Service,
            that will also prevent deletions or re-creates of Feature Views used in that service. `prevent_destroy` is
            only enforced in live (i.e. non-dev) workspaces.
        :param batch_config: BatchConfig object containing the configuration of the Batch Data Source that backs this
            Tecton Stream Source. This field is optional only if `stream_config` is a PushConfig.
        :param stream_config: StreamConfig object containing the configuration of the
            Stream Data Source that backs this Tecton Stream Source.
        :param options: Additional options to configure the Source. Used for advanced use cases and beta features.
        :param schema: A schema for the StreamSource. If not provided, the schema will be inferred from the underlying batch source.
            Right now, schemas can only be specified for StreamSources with a PushConfig, and that's also why the schema must be a list of Tecton types.
        """
        schema_container = (
            schema_container_pb2.SchemaContainer(tecton_schema=type_utils.to_tecton_schema(schema)) if schema else None
        )

        data_source_type: data_source_type_pb2.DataSourceType
        if isinstance(stream_config, configs.PushConfig):
            if batch_config:
                data_source_type = data_source_type_pb2.DataSourceType.PUSH_WITH_BATCH
            else:
                data_source_type = data_source_type_pb2.DataSourceType.PUSH_NO_BATCH
        else:
            assert batch_config is not None, f"batch_config must be provided for stream source {name}"
            data_source_type = data_source_type_pb2.DataSourceType.STREAM_WITH_BATCH

        ds_args = virtual_data_source__args_pb2.VirtualDataSourceArgs(
            virtual_data_source_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            version=framework_version_pb2.FrameworkVersion.FWV5,
            type=data_source_type,
            prevent_destroy=prevent_destroy,
            options=options,
        )
        if schema_container:
            ds_args.schema.CopyFrom(schema_container)
        if batch_config:
            batch_config._merge_batch_args(ds_args)
        stream_config._merge_stream_args(ds_args)
        info = base_tecton_object.TectonObjectInfo.from_args_proto(ds_args.info, ds_args.virtual_data_source_id)
        source_info = construct_fco_source_info(ds_args.virtual_data_source_id)

        self.__attrs_init__(
            info=info,
            spec=None,
            args=ds_args,
            source_info=source_info,
            args_supplement=_build_args_supplement(batch_config, stream_config),
        )

        if not conf.get_bool("TECTON_SKIP_OBJECT_VALIDATION"):
            self._validate()
        self._populate_spec()

        base_tecton_object._register_local_object(self)

    @classmethod
    @typechecked
    def _from_spec(cls, spec: specs.DataSourceSpec) -> "StreamSource":
        """Create a StreamSource from directly from a spec. Specs are assumed valid and will not be re-validated."""
        info = base_tecton_object.TectonObjectInfo.from_spec(spec)
        obj = cls.__new__(cls)  # Instantiate the object. Does not call init.
        obj.__attrs_init__(
            info=info,
            spec=spec,
            args=None,
            source_info=None,
            args_supplement=None,
        )
        return obj

    @sdk_decorators.sdk_public_method
    def start_stream_preview(
        self,
        table_name: str,
        *,
        apply_translator: bool = True,
        option_overrides: Optional[Dict[str, str]] = None,
        checkpoint_dir: Optional[str] = None,
    ) -> pyspark_streaming.StreamingQuery:
        """
        Starts a streaming job to write incoming records from this DS's stream to a temporary table with a given name.

        After records have been written to the table, they can be queried using ``spark.sql()``. If ran in a Databricks
        notebook, Databricks will also automatically visualize the number of incoming records.

        This is a testing method, most commonly used to verify a StreamDataSource is correctly receiving streaming events.
        Note that the table will grow infinitely large, so this is only really useful for debugging in notebooks.

        :param table_name: The name of the temporary table that this method will write to.
        :param apply_translator: Whether to apply this data source's ``raw_stream_translator``.
            When True, the translated data will be written to the table. When False, the
            raw, untranslated data will be written. ``apply_translator`` is not applicable to stream sources configured
            with ``spark_stream_config`` because it does not have a ``post_processor``.
        :param option_overrides: A dictionary of Spark readStream options that will override any readStream options set
            by the data source. Can be used to configure behavior only for the preview, e.g. setting
            ``startingOffsets:latest`` to preview only the most recent events in a Kafka stream.
        :param checkpoint_dir: A root directory that the streaming job will checkpoint to.
        """
        if is_stream_ingest_data_source(self._spec.type):
            msg = "Cannot preview stream ingest data sources"
            raise ValueError(msg)
        return spark_api.start_stream_preview(
            self._spec, table_name, apply_translator, option_overrides, checkpoint_dir
        )

    @sdk_decorators.sdk_public_method
    @sdk_decorators.assert_remote_object
    def ingest(self, events: Union[Dict[str, Any], Sequence[Dict[str, Any]]], dry_run: bool = False) -> Dict[str, Any]:
        """Ingests a single event into the Tecton Online Ingest API.

        :param events: A list of dictionaries representing a sequence of events to be ingested. Also accepts a
            single dictionary.
        :param dry_run: If True, the ingest request will be validated, but the event will not be materialized.
            If False, the event will be materialized.
        """
        if not is_stream_ingest_data_source(self._spec.type):
            msg = "Can only ingest events for stream sources with push configs"
            raise ValueError(msg)
        if not self._spec:
            msg = "Cannot ingest events for a stream source that has not been applied"
            raise ValueError(msg)

        if isinstance(events, dict):
            events = [events]

        status_code, reason, response = IngestionClient().ingest(
            workspace_name=self._spec.workspace,
            push_source_name=self._spec.name,
            ingestion_records=events,
            dry_run=dry_run,
        )
        if status_code >= 500:
            raise errors.INTERNAL_ERROR(message=json.dumps(response))
        elif status_code >= 400:
            raise errors.INGESTAPI_USER_ERROR(
                status_code=status_code, reason=reason, error_message=json.dumps(response)
            )
        else:
            return response


@attrs.define(eq=False)
class PushSource(StreamSource):
    """Deprecated. Use StreamSource with a PushConfig instead.
    A Tecton PushSource, used to configure the Tecton Online Ingest API for use in a StreamFeatureView.

    ``PushSource`` is currently in private preview, please contact Tecton support if you are interested in participating in the preview.

    A PushSource may also contain an optional batch config for backfilling and offline training data generation.

    Example of a PushSource declaration:

    .. code-block:: python

        from tecton import HiveConfig, PushSource, BatchSource
        from tecton.types import Field, Int64, String, Timestamp

        # Declare a schema for the Push Source
        input_schema = [
            Field(name='user_id', dtype=String),
            Field(name='event_timestamp', dtype=String),
            Field(name='clicked', dtype=Int64),
        ]

        # Declare a PushSource with a name, schema and a batch_config parameters
        # See the API documentation for BatchConfig
        click_event_source = PushSource(
                                name="click_event_source",
                                schema=input_schema,
                                batch_config=HiveConfig(
                                    database='demo_ads',
                                    table='impressions_batch',
                                ),
                                description="Sample Push Source for click events",
                                owner="pooja@tecton.ai",
                                tags={'release': 'staging'}
                                )
    """

    @typechecked
    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = None,
        prevent_destroy: bool = False,
        schema: List[types.Field],
        batch_config: Optional[BatchConfigType] = None,
        options: Optional[Dict[str, str]] = None,
    ):
        """Creates a new PushSource.

        :param name: A unique name of the DataSource.
        :param schema: A schema for the PushSource
        :param description: A human-readable description.
        :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
        :param owner: Owner name (typically the email of the primary maintainer).
        :param prevent_destroy: If True, this Tecton object will be blocked from being deleted or re-created (i.e. a
            destructive update) during tecton plan/apply. To remove or update this object, `prevent_destroy` must be
            set to False via the same tecton apply or a separate tecton apply. `prevent_destroy` can be used to prevent accidental changes
            such as inadvertantly deleting a Feature Service used in production or recreating a Feature View that
            triggers expensive rematerialization jobs. `prevent_destroy` also blocks changes to dependent Tecton objects
            that would trigger a recreate of the tagged object, e.g. if `prevent_destroy` is set on a Feature Service,
            that will also prevent deletions or re-creates of Feature Views used in that service. `prevent_destroy` is
            only enforced in live (i.e. non-dev) workspaces.
        :param batch_config: An optional BatchConfig object containing the configuration of the Batch Data Source that backs
            this Tecton Push Source. The Batch Source's schema must contain a super-set of all the columns defined in the Push Source schema.
        :param options: Additional options to configure the Source. Used for advanced use cases and beta features.
        """
        logger.warning("PushSource is deprecated. Please use StreamSource with a PushConfig instead")

        data_source_type = (
            data_source_type_pb2.DataSourceType.PUSH_WITH_BATCH
            if batch_config is not None
            else data_source_type_pb2.DataSourceType.PUSH_NO_BATCH
        )

        ds_args = virtual_data_source__args_pb2.VirtualDataSourceArgs(
            virtual_data_source_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            version=framework_version_pb2.FrameworkVersion.FWV5,
            type=data_source_type,
            prevent_destroy=prevent_destroy,
            schema=schema_container_pb2.SchemaContainer(tecton_schema=type_utils.to_tecton_schema(schema)),
            options=options,
        )
        if batch_config is not None:
            batch_config._merge_batch_args(ds_args)
        info = base_tecton_object.TectonObjectInfo.from_args_proto(ds_args.info, ds_args.virtual_data_source_id)
        source_info = construct_fco_source_info(ds_args.virtual_data_source_id)

        self.__attrs_init__(
            info=info,
            spec=None,
            args=ds_args,
            source_info=source_info,
            args_supplement=_build_args_supplement(batch_config, None),
        )

        if not conf.get_bool("TECTON_SKIP_OBJECT_VALIDATION"):
            self._validate()
        self._populate_spec()

        base_tecton_object._register_local_object(self)


def data_source_from_spec(data_source_spec: specs.DataSourceSpec):
    """Create a Data Source (of the correct type) from the provided spec."""
    if data_source_spec.type in (
        data_source_type_pb2.DataSourceType.STREAM_WITH_BATCH,
        data_source_type_pb2.DataSourceType.PUSH_WITH_BATCH,
        data_source_type_pb2.DataSourceType.PUSH_NO_BATCH,
    ):
        return StreamSource._from_spec(data_source_spec)
    elif data_source_spec.type == data_source_type_pb2.DataSourceType.BATCH:
        return BatchSource._from_spec(data_source_spec)
    else:
        msg = f"Unexpected Data Source Type. Spec: {data_source_spec}"
        raise ValueError(msg)


def _build_args_supplement(
    batch_config: Optional[BatchConfigType], stream_config: Optional[StreamConfigType]
) -> specs.DataSourceSpecArgsSupplement:
    supplement = specs.DataSourceSpecArgsSupplement()
    if isinstance(
        batch_config,
        (
            configs.FileConfig,
            configs.HiveConfig,
            configs.RedshiftConfig,
            configs.SnowflakeConfig,
            configs.UnityConfig,
            configs.BigQueryConfig,
        ),
    ):
        supplement.batch_post_processor = batch_config.post_processor
    elif isinstance(batch_config, (configs.PandasBatchConfig, configs.SparkBatchConfig)):
        supplement.batch_data_source_function = batch_config.data_source_function
    elif batch_config is not None:
        msg = f"Unexpected batch_config type: {batch_config}"
        raise TypeError(msg)

    if isinstance(stream_config, (configs.KinesisConfig, configs.KafkaConfig)):
        supplement.stream_post_processor = stream_config.post_processor
    elif isinstance(stream_config, configs.SparkStreamConfig):
        supplement.stream_data_source_function = stream_config.data_source_function
    elif isinstance(stream_config, configs.PushConfig):
        pass
    elif stream_config is not None:
        msg = f"Unexpected stream_config type: {stream_config}"
        raise TypeError(msg)

    return supplement


def is_stream_ingest_data_source(data_source_type: data_source_type_pb2.DataSourceType) -> bool:
    return data_source_type in {
        data_source_type_pb2.DataSourceType.PUSH_WITH_BATCH,
        data_source_type_pb2.DataSourceType.PUSH_NO_BATCH,
    }



from tecton_proto.auth import authorization_service__client_pb2 as authorization_service_pb2
from tecton_proto.materializationjobservice import (
    materialization_job_service__client_pb2 as materialization_job_service_pb2,
)
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2
from tecton_proto.modelartifactservice import model_artifact_service__client_pb2 as model_artifact_service_pb2
from tecton_proto.remoteenvironmentservice import (
    remote_environment_service__client_pb2 as remote_environment_service_pb2,
)
from tecton_proto.secrets import secrets_service__client_pb2 as secrets_service_client_pb2
from tecton_proto.testhelperservice import test_helper_service__client_pb2 as test_helper_service_pb2


GRPC_SERVICE_MODULES = [
    metadata_service_pb2,
    materialization_job_service_pb2,
    authorization_service_pb2,
    remote_environment_service_pb2,
    test_helper_service_pb2,
    secrets_service_client_pb2,
    model_artifact_service_pb2,
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/api/featureservice/feature_service_request__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nEtecton_proto/api/featureservice/feature_service_request__client.proto\x12 tecton_proto.api.feature_service\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xa5\x02\n\x0fMetadataOptions\x12#\n\rinclude_names\x18\x01 \x01(\x08R\x0cincludeNames\x12\x36\n\x17include_effective_times\x18\x02 \x01(\x08R\x15includeEffectiveTimes\x12\'\n\rinclude_types\x18\x03 \x01(\x08\x42\x02\x18\x01R\x0cincludeTypes\x12,\n\x12include_data_types\x18\x05 \x01(\x08R\x10includeDataTypes\x12(\n\x10include_slo_info\x18\x04 \x01(\x08R\x0eincludeSloInfo\x12\x34\n\x16include_serving_status\x18\x06 \x01(\x08R\x14includeServingStatus\"\xbe\x01\n\x0eRequestOptions\x12,\n\x0fread_from_cache\x18\x01 \x01(\x08:\x04trueR\rreadFromCache\x12*\n\x0ewrite_to_cache\x18\x02 \x01(\x08:\x04trueR\x0cwriteToCache\x12R\n\x17high_watermark_override\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x15highWatermarkOverride\"e\n\x12GetFeaturesRequest\x12O\n\x06params\x18\x01 \x01(\x0b\x32\x37.tecton_proto.api.feature_service.GetFeaturesParametersR\x06params\"o\n\x17GetFeaturesBatchRequest\x12T\n\x06params\x18\x01 \x01(\x0b\x32<.tecton_proto.api.feature_service.GetFeaturesBatchParametersR\x06params\"\xba\x08\n\x15GetFeaturesParameters\x12.\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\tH\x00R\x10\x66\x65\x61tureServiceId\x12\x32\n\x14\x66\x65\x61ture_service_name\x18\x02 \x01(\tH\x00R\x12\x66\x65\x61tureServiceName\x12.\n\x12\x66\x65\x61ture_package_id\x18\x0c \x01(\tH\x00R\x10\x66\x65\x61turePackageId\x12\x32\n\x14\x66\x65\x61ture_package_name\x18\r \x01(\tH\x00R\x12\x66\x65\x61turePackageName\x12(\n\x0f\x66\x65\x61ture_view_id\x18\x0f \x01(\tH\x00R\rfeatureViewId\x12,\n\x11\x66\x65\x61ture_view_name\x18\x10 \x01(\tH\x00R\x0f\x66\x65\x61tureViewName\x12%\n\x0eworkspace_name\x18\n \x01(\tR\rworkspaceName\x12i\n\x0cjoin_key_map\x18\x08 \x03(\x0b\x32G.tecton_proto.api.feature_service.GetFeaturesParameters.JoinKeyMapEntryR\njoinKeyMap\x12~\n\x13request_context_map\x18\t \x03(\x0b\x32N.tecton_proto.api.feature_service.GetFeaturesParameters.RequestContextMapEntryR\x11requestContextMap\x12\\\n\x10metadata_options\x18\x07 \x01(\x0b\x32\x31.tecton_proto.api.feature_service.MetadataOptionsR\x0fmetadataOptions\x12\x32\n\x15\x61llow_partial_results\x18\x0b \x01(\x08R\x13\x61llowPartialResults\x12Y\n\x0frequest_options\x18\x12 \x01(\x0b\x32\x30.tecton_proto.api.feature_service.RequestOptionsR\x0erequestOptions\x12$\n\risCallerBatch\x18\x11 \x01(\x08R\risCallerBatch\x1aU\n\x0fJoinKeyMapEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x1a\\\n\x16RequestContextMapEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x42\t\n\x07locatorJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07J\x04\x08\x0e\x10\x0f\"\xf2\x02\n\x1aGetFeaturesBatchParameters\x12.\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\tH\x00R\x10\x66\x65\x61tureServiceId\x12\x32\n\x14\x66\x65\x61ture_service_name\x18\x02 \x01(\tH\x00R\x12\x66\x65\x61tureServiceName\x12%\n\x0eworkspace_name\x18\x03 \x01(\tR\rworkspaceName\x12`\n\x0crequest_data\x18\x04 \x03(\x0b\x32=.tecton_proto.api.feature_service.GetFeaturesBatchRequestDataR\x0brequestData\x12\\\n\x10metadata_options\x18\x05 \x01(\x0b\x32\x31.tecton_proto.api.feature_service.MetadataOptionsR\x0fmetadataOptionsB\t\n\x07locator\"\xca\x03\n\x1bGetFeaturesBatchRequestData\x12o\n\x0cjoin_key_map\x18\x01 \x03(\x0b\x32M.tecton_proto.api.feature_service.GetFeaturesBatchRequestData.JoinKeyMapEntryR\njoinKeyMap\x12\x84\x01\n\x13request_context_map\x18\x02 \x03(\x0b\x32T.tecton_proto.api.feature_service.GetFeaturesBatchRequestData.RequestContextMapEntryR\x11requestContextMap\x1aU\n\x0fJoinKeyMapEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x1a\\\n\x16RequestContextMapEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x05value:\x02\x38\x01\x42\x36Z4github.com/tecton-ai/tecton_proto/api/featureservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.api.featureservice.feature_service_request__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z4github.com/tecton-ai/tecton_proto/api/featureservice'
  _METADATAOPTIONS.fields_by_name['include_types']._options = None
  _METADATAOPTIONS.fields_by_name['include_types']._serialized_options = b'\030\001'
  _GETFEATURESPARAMETERS_JOINKEYMAPENTRY._options = None
  _GETFEATURESPARAMETERS_JOINKEYMAPENTRY._serialized_options = b'8\001'
  _GETFEATURESPARAMETERS_REQUESTCONTEXTMAPENTRY._options = None
  _GETFEATURESPARAMETERS_REQUESTCONTEXTMAPENTRY._serialized_options = b'8\001'
  _GETFEATURESBATCHREQUESTDATA_JOINKEYMAPENTRY._options = None
  _GETFEATURESBATCHREQUESTDATA_JOINKEYMAPENTRY._serialized_options = b'8\001'
  _GETFEATURESBATCHREQUESTDATA_REQUESTCONTEXTMAPENTRY._options = None
  _GETFEATURESBATCHREQUESTDATA_REQUESTCONTEXTMAPENTRY._serialized_options = b'8\001'
  _METADATAOPTIONS._serialized_start=171
  _METADATAOPTIONS._serialized_end=464
  _REQUESTOPTIONS._serialized_start=467
  _REQUESTOPTIONS._serialized_end=657
  _GETFEATURESREQUEST._serialized_start=659
  _GETFEATURESREQUEST._serialized_end=760
  _GETFEATURESBATCHREQUEST._serialized_start=762
  _GETFEATURESBATCHREQUEST._serialized_end=873
  _GETFEATURESPARAMETERS._serialized_start=876
  _GETFEATURESPARAMETERS._serialized_end=1958
  _GETFEATURESPARAMETERS_JOINKEYMAPENTRY._serialized_start=1738
  _GETFEATURESPARAMETERS_JOINKEYMAPENTRY._serialized_end=1823
  _GETFEATURESPARAMETERS_REQUESTCONTEXTMAPENTRY._serialized_start=1825
  _GETFEATURESPARAMETERS_REQUESTCONTEXTMAPENTRY._serialized_end=1917
  _GETFEATURESBATCHPARAMETERS._serialized_start=1961
  _GETFEATURESBATCHPARAMETERS._serialized_end=2331
  _GETFEATURESBATCHREQUESTDATA._serialized_start=2334
  _GETFEATURESBATCHREQUESTDATA._serialized_end=2792
  _GETFEATURESBATCHREQUESTDATA_JOINKEYMAPENTRY._serialized_start=1738
  _GETFEATURESBATCHREQUESTDATA_JOINKEYMAPENTRY._serialized_end=1823
  _GETFEATURESBATCHREQUESTDATA_REQUESTCONTEXTMAPENTRY._serialized_start=1825
  _GETFEATURESBATCHREQUESTDATA_REQUESTCONTEXTMAPENTRY._serialized_end=1917
# @@protoc_insertion_point(module_scope)
